r"""Definition of common enumerations."""

from __future__ import annotations

from enum import Enum, IntEnum
import itertools
import re
from typing import Iterable, List, Optional, Sequence, Set, Tuple, Union, overload


class Sampling(Enum):
    r"""Enumeration of image interpolation modes."""

    AREA = "area"
    BICUBIC = "bicubic"
    BSPLINE = "bspline"  # cubic B-spline
    LINEAR = "linear"  # bilinear or trilinear
    NEAREST = "nearest"

    @classmethod
    def from_arg(cls, arg: Union[Sampling, str, None]) -> Sampling:
        r"""Create enumeration value from function argument."""
        if isinstance(arg, str):
            arg = arg.lower()
        if arg is None or arg in ("default", "bilinear", "trilinear"):
            return cls.LINEAR
        if arg == "nn":
            return cls.NEAREST
        return cls(arg)

    def grid_sample_mode(self, num_spatial_dim: int) -> str:
        r"""Interpolation mode argument for torch.nn.functional.grid_sample() for given number of spatial dimensions."""
        if self == self.LINEAR:
            return "bilinear"
        if self == self.NEAREST:
            return "nearest"
        raise ValueError(
            f"torch.nn.functional.grid_sample() does not support padding mode '{self.value}' for {num_spatial_dim}-dimensional images"
        )

    def interpolate_mode(self, num_spatial_dim: int) -> str:
        r"""Interpolation mode argument for torch.nn.functional.interpolate() for given number of spatial dimensions."""
        if self == self.AREA:
            return "area"
        if self == self.BICUBIC:
            return "bicubic"
        if self == self.LINEAR:
            if num_spatial_dim == 1:
                return "linear"
            if num_spatial_dim == 2:
                return "bilinear"
            if num_spatial_dim == 3:
                return "trilinear"
        if self == self.NEAREST:
            return "nearest"
        raise ValueError(
            f"torch.nn.functional.interpolate() does not support padding mode '{self.value}' for {num_spatial_dim}-dimensional images"
        )


class PaddingMode(Enum):
    r"""Enumeration of image extrapolation modes."""

    NONE = "none"
    CONSTANT = "constant"
    BORDER = "border"
    REFLECT = "reflect"
    REPLICATE = "replicate"
    ZEROS = "zeros"

    @classmethod
    def from_arg(cls, arg: Union[PaddingMode, str, None]) -> PaddingMode:
        r"""Create enumeration value from function argument."""
        if isinstance(arg, str):
            arg = arg.lower()
        if arg is None or arg == "default":
            return cls.ZEROS
        if arg in ("mirror", "reflection"):
            return cls.REFLECT
        if arg == "circular":
            return cls.REPLICATE
        return cls(arg)

    def conv_mode(self, num_spatial_dim: int = 3) -> str:
        r"""Padding mode argument for torch.nn.ConvNd()."""
        if self in (self.CONSTANT, self.ZEROS):
            return "zeros"
        elif self == self.REFLECT:
            return "reflect"
        elif self == self.REPLICATE:
            return "replicate"
        raise ValueError(
            f"torch.nn.Conv{num_spatial_dim}d() does not support padding mode '{self.value}'"
        )

    def grid_sample_mode(self, num_spatial_dim: int) -> str:
        r"""Padding mode argument for torch.nn.functional.grid_sample()."""
        if 2 <= num_spatial_dim <= 3:
            if self in (self.CONSTANT, self.ZEROS):
                return "zeros"
            if self == self.BORDER:
                return "border"
            if self == self.REFLECT:
                return "reflection"
        raise ValueError(
            f"torch.nn.functional.grid_sample() does not support padding mode '{self.value}' for {num_spatial_dim}-dimensional images"
        )

    def pad_mode(self, num_spatial_dim: int) -> str:
        r"""Padding mode argument for torch.nn.functional.pad() for given number of spatial dimensions."""
        if self == self.CONSTANT:
            return "constant"
        elif self == self.REFLECT:
            if 1 <= num_spatial_dim <= 2:
                return "reflect"
        elif self == self.REPLICATE:
            if 1 <= num_spatial_dim <= 3:
                return "replicate"
        raise ValueError(
            f"torch.nn.functional.pad() does not support padding mode '{self.value}' for {num_spatial_dim}-dimensional images"
        )


class SpatialDim(IntEnum):
    r"""Spatial image dimension selector."""

    X = 0
    Y = 1
    Z = 2
    T = 3

    @classmethod
    def from_arg(cls, arg: Union[int, str, SpatialDim]) -> SpatialDim:
        r"""Get enumeration value from function argument."""
        if arg in ("x", "X"):
            return cls.X
        if arg in ("y", "Y"):
            return cls.Y
        if arg in ("z", "Z"):
            return cls.Z
        if arg in ("t", "T"):
            return cls.T
        return cls(arg)

    def symbol(self) -> str:
        r"""Letter of spatial dimension."""
        return ("x", "y", "z", "t")[self.value]

    def tensor_dim(self, ndim: int, channels_last: bool = False) -> int:
        r"""Map spatial dimension identifier to image data tensor dimension."""
        dim = ndim - (2 if channels_last else 1) - self.value
        if (
            channels_last
            and (dim < 1 or dim > ndim - 2)
            or (not channels_last and (dim < 2 or dim > ndim - 1))
        ):
            raise ValueError("SpatialDim.tensor_dim() is out-of-bounds")
        return dim

    def __str__(self) -> str:
        r"""Letter of spatial dimension."""
        return self.symbol()


SpatialDimArg = Union[int, str, SpatialDim]

SpatialDerivativeKey = str


class SpatialDerivativeKeys(object):
    r"""Auxiliary functions for identifying and enumerating spatial derivatives.

    Spatial derivatives are encoded by a sequence of letters, where each letter
    identifies the spatial dimension (cf. ``SpatialDim``) along which to take the
    derivative. The length of the string encoding determines the order of the
    derivative, i.e., how many times the input image is being derived with
    respect to one or more spatial dimensions.

    """

    @staticmethod
    def check(arg: Union[str, Iterable[str]]):
        r"""Check if given derivatives key is valid."""
        if isinstance(arg, str):
            arg = (arg,)
        for key in arg:
            if not isinstance(key, str):
                raise TypeError("Spatial derivatives key must be str")
            if re.search(r"[^xyzt]", key):
                raise ValueError(
                    "Spatial derivatives key must only contain letters 'x', 'y', 'z', or 't'"
                )

    @classmethod
    def is_valid(cls, arg: Union[str, Iterable[str]]) -> bool:
        r"""Check if given derivatives key is valid."""
        try:
            cls.check(arg)
        except (TypeError, ValueError):
            return False
        return True

    @staticmethod
    def is_mixed(key: SpatialDerivativeKey) -> bool:
        r"""Whether derivative contains mixed terms."""
        return len(set(key)) > 1

    @staticmethod
    def all(ndim: int, order: Union[int, Sequence[int]]) -> List[SpatialDerivativeKey]:
        r"""Unmixed spatial derivatives of specified order."""
        if isinstance(order, int):
            order = [order]
        keys = []
        dims = [str(SpatialDim(d)) for d in range(ndim)]
        for n in order:
            if n > 0:
                codes = dims
                for _ in range(1, n):
                    codes = [code + letter for code, letter in itertools.product(codes, dims)]
                keys.extend(codes)
        return keys

    @staticmethod
    def unmixed(ndim: int, order: int) -> List[SpatialDerivativeKey]:
        r"""Unmixed spatial derivatives of specified order."""
        if order <= 0:
            return []
        return [SpatialDim(d).symbol() * order for d in range(ndim)]

    @classmethod
    def unique(cls, keys: Iterable[SpatialDerivativeKey]) -> Set[SpatialDerivativeKey]:
        r"""Unique spatial derivatives."""
        return set((cls.sorted(key) for key in keys))

    @classmethod
    def sorted(cls, key: SpatialDerivativeKey) -> SpatialDerivativeKey:
        r"""Sort letters of spatial dimensions in spatial derivative key."""
        return cls.join(sorted(cls.split(key)))

    @staticmethod
    def order(arg: SpatialDerivativeKey) -> int:
        r"""Order of the spatial derivative."""
        return len(arg)

    @classmethod
    def max_order(cls, keys: Iterable[SpatialDerivativeKey]) -> int:
        return max([0] + [cls.order(key) for key in keys])

    @staticmethod
    def split(arg: SpatialDerivativeKey) -> List[SpatialDim]:
        r"""Split spatial derivative key into spatial dimensions enum values."""
        return [SpatialDim.from_arg(letter) for letter in arg]

    @staticmethod
    def join(arg: Sequence[SpatialDim]) -> SpatialDerivativeKey:
        r"""Join spatial dimensions to spatial derivative key."""
        return "".join(sdim.symbol() for sdim in arg)


class FlowChannelIndex(IntEnum):
    r"""Flow vector component index."""

    U = 0
    V = 1
    W = 2

    @classmethod
    def from_arg(cls, arg: Union[int, str, FlowChannelIndex]) -> FlowChannelIndex:
        r"""Get enumeration value from function argument."""
        if arg in ("u", "U"):
            return cls.U
        if arg in ("v", "V"):
            return cls.V
        if arg in ("w", "W"):
            return cls.W
        return cls(arg)

    def index(self) -> int:
        r"""Map flow component identifier to tensor channel index."""
        return self.value

    def symbol(self) -> str:
        r"""Single lowercase letter identifying this flow component."""
        return ("u", "v", "w")[self.value]

    def __str__(self) -> str:
        return self.symbol()


FlowChannelIndexArg = Union[int, str, FlowChannelIndex]

FlowDerivativeKey = str


class FlowDerivativeKeys(object):
    r"""Auxiliary functions for identifying and enumerating spatial derivatives of a flow field.

    The flow field components are denoted as "u" (c=0), "v"  (c=1), and "w"  (c=2), respectively,
    where c is the channel index. The spatial dimensions along which to take derivatives are
    identified by the letters "x", "y", and "z" (cf. :class:`SpatialDerivativeKeys`). The partial
    derivative of a given vector field component along one or more spatial dimensions is denoted
    by a quotient string such as "dv/dz" or "du/dxy". When multiple flow field components are
    specified, the spatial derivative of each is computed, i.e., "duw/dx" is shorthand for
    ["du/dx", "dw/dx"].

    """

    @classmethod
    def from_arg(
        cls,
        spatial_dims: int,
        which: Optional[Union[str, Sequence[str]]] = None,
        order: Optional[int] = None,
    ) -> List[FlowDerivativeKey]:
        r"""Get flow derivative keys from function arguments.

        See also ``which`` parameter of :func:`flow_derivatives()` function.

        Args:
            spatial_dims: Number of spatial dimensions.
            which: String codes of spatial deriviatives to compute. When only a sequence of spatial
                dimension keys is given (cf. :class:`SpatialDerivateKeys`), the respective spatial
                derivative is computed for all vector field components, i.e., for ``spatial_dims=3``,
                "x" is shorthand for "du/dx", "dv/dx", and "dw/dx".
            order: Order of spatial derivatives. When both ``which`` and ``order`` are specified,
                only the flow field derivatives listed in ``which`` and are of the given order
                are returned.

        Returns:
            Flow derivative keys, one for each partial derivative scalar field.

        """
        if spatial_dims < 2 or spatial_dims > 3:
            raise ValueError("Spatial flow field dimensions must be 2 or 3")
        if which is None:
            if order is None:
                order = 1
            keys = cls.all(spatial_dims=spatial_dims, order=order)
        else:
            keys = []
            regex = re.compile(r"^(d(?P<channels>[uvw]+)/d)?(?P<derivs>[xyzt]+)$")
            if isinstance(which, str):
                which = [which]
            for arg in which:
                match = regex.match(arg)
                if not match:
                    raise ValueError(f"Invalid flow derivative specification: {arg}")
                derivs = match["derivs"]
                SpatialDerivativeKeys.check(derivs)
                if order is None or SpatialDerivativeKeys.order(derivs) == order:
                    if match["channels"] is None:
                        channels = ["u", "v", "w"][:spatial_dims]
                    else:
                        channels = [str(c).lower() for c in match["channels"]]
                    for channel in channels:
                        keys.append(f"d{channel}/d{derivs}")
        return keys

    @overload
    @classmethod
    def unique(cls, arg: FlowDerivativeKey) -> FlowDerivativeKey:
        r"""Unique spatial flow derivative key, where spatial dimension codes are sorted alphabetically."""
        ...

    @overload
    @classmethod
    def unique(cls, arg: Iterable[FlowDerivativeKey]) -> Set[FlowDerivativeKey]:
        r"""Set of unique spatial flow derivative keys."""
        ...

    @classmethod
    def unique(
        cls, arg: Union[FlowDerivativeKey, Iterable[FlowDerivativeKey]]
    ) -> Union[FlowDerivativeKey, Set[FlowDerivativeKey]]:
        r"""Unique spatial flow derivative."""
        unique_keys = set()
        keys = [arg] if isinstance(arg, str) else arg
        for channel, derivative_key in cls.split(keys):
            derivative_key = SpatialDerivativeKeys.sorted(derivative_key)
            unique_key = cls.symbol(channel, derivative_key)
            unique_keys.add(unique_key)
        if arg is not keys:
            return next(iter(unique_keys))
        return unique_keys

    @classmethod
    def sorted(cls, keys: Iterable[FlowDerivativeKey]) -> List[FlowDerivativeKey]:
        r"""Ordered list of flow derivative keys."""
        return sorted(keys)

    @classmethod
    def is_mixed(cls, arg: FlowDerivativeKey) -> bool:
        r"""Whether flow derivative is taken along more than one spatial dimension."""
        return SpatialDerivativeKeys.is_mixed(cls.split(arg)[1])

    @classmethod
    def order(cls, arg: FlowDerivativeKey) -> int:
        r"""Order of the spatial flow derivative."""
        return SpatialDerivativeKeys.order(cls.split(arg)[1])

    @classmethod
    def max_order(cls, keys: Iterable[SpatialDerivativeKey]) -> int:
        return SpatialDerivativeKeys.max_order(derivs for _, derivs in cls.split(keys))

    @overload
    @staticmethod
    def split(arg: FlowDerivativeKey) -> Tuple[FlowChannelIndex, SpatialDerivativeKey]:
        ...

    @overload
    @staticmethod
    def split(
        arg: Iterable[FlowDerivativeKey],
    ) -> List[Tuple[FlowChannelIndex, SpatialDerivativeKey]]:
        ...

    @staticmethod
    def split(
        arg: Union[FlowDerivativeKey, Iterable[FlowDerivativeKey]]
    ) -> List[Tuple[FlowChannelIndex, SpatialDerivativeKey]]:
        r"""Split flow derivative keys into flow channel index and spatial derivative key."""
        keys = []
        regex = re.compile(r"^d(?P<channel_key>[uvw])/d(?P<derivative_key>[xXyYzZtT]+)$")
        args = [arg] if isinstance(arg, str) else arg
        for key in args:
            match = regex.match(key)
            if not match:
                raise ValueError(f"Invalid flow derivative key: {key}")
            channel_key = match["channel_key"]
            derivative_key = match["derivative_key"]
            channel_index = FlowChannelIndex.from_arg(channel_key)
            SpatialDerivativeKeys.check(derivative_key)
            keys.append((channel_index, derivative_key))
        if arg is not args:
            return keys[0]
        return keys

    @staticmethod
    def symbol(
        channel: FlowChannelIndexArg,
        *args: Union[str, int, SpatialDim],
    ) -> FlowDerivativeKey:
        channel = FlowChannelIndex.from_arg(channel)
        derivs = []
        for arg in args:
            if isinstance(arg, str):
                derivs.extend(SpatialDerivativeKeys.split(arg))
            elif isinstance(arg, int):
                derivs.append(SpatialDim(arg))
            elif isinstance(arg, SpatialDim):
                derivs.append(arg)
            else:
                raise TypeError("Spatial derivative identifier must be int, str, or SpatialDim")
        return f"d{channel.symbol()}/d{SpatialDerivativeKeys.join(derivs)}"

    @classmethod
    def all(
        cls,
        spatial_dims: int,
        channel: Optional[Union[FlowChannelIndexArg, Sequence[FlowChannelIndexArg]]] = None,
        order: Union[int, Sequence[int]] = 1,
    ) -> List[FlowDerivativeKey]:
        r"""All spatial derivatives of specified order."""
        if order < 0:
            raise ValueError("Spatial derivatives order must be non-negative")
        if order == 0:
            return []
        channels = cls._channels(spatial_dims, channel)
        derivs = SpatialDerivativeKeys.all(spatial_dims, order=order)
        return [cls.symbol(c, d) for c, d in itertools.product(channels, derivs)]

    @classmethod
    def unmixed(
        cls,
        spatial_dims: int,
        channel: Optional[Union[FlowChannelIndexArg, Sequence[FlowChannelIndexArg]]] = None,
        order: Union[int, Sequence[int]] = 1,
    ) -> List[FlowDerivativeKey]:
        r"""Unmixed spatial derivatives of specified order."""
        if order < 0:
            raise ValueError("Spatial derivatives order must be non-negative")
        if order == 0:
            return []
        channels = cls._channels(spatial_dims, channel)
        derivs = SpatialDerivativeKeys.unmixed(spatial_dims, order=order)
        return [cls.symbol(c, d) for c, d in itertools.product(channels, derivs)]

    @classmethod
    def gradient(
        cls,
        spatial_dims: int,
        channel: Optional[Union[FlowChannelIndexArg, Sequence[FlowChannelIndexArg]]] = None,
    ) -> List[FlowDerivativeKey]:
        return cls.all(spatial_dims=spatial_dims, channel=channel, order=1)

    @classmethod
    def jacobian(cls, spatial_dims: int) -> List[FlowDerivativeKey]:
        return cls.all(spatial_dims=spatial_dims, order=1)

    @classmethod
    def divergence(cls, spatial_dims: int) -> List[FlowDerivativeKey]:
        return [cls.symbol(i, i) for i in range(spatial_dims)]

    @classmethod
    def curvature(cls, spatial_dims: int) -> List[FlowDerivativeKey]:
        channels = range(spatial_dims)
        derivs = SpatialDerivativeKeys.unmixed(spatial_dims, order=2)
        return [cls.symbol(c, d) for c, d in itertools.product(channels, derivs)]

    @classmethod
    def hessian(
        cls,
        spatial_dims: int,
        channel: Optional[Union[FlowChannelIndexArg, Sequence[FlowChannelIndexArg]]] = None,
    ) -> List[FlowDerivativeKey]:
        return cls.all(spatial_dims=spatial_dims, channel=channel, order=2)

    @staticmethod
    def _channels(
        spatial_dims: int,
        channel: Optional[Union[FlowChannelIndexArg, Sequence[FlowChannelIndexArg]]] = None,
    ) -> List[FlowChannelIndex]:
        if channel is None:
            channel = range(spatial_dims)
        elif isinstance(channel, (int, str, FlowChannelIndex)):
            channel = [channel]
        return [FlowChannelIndex.from_arg(c) for c in channel]



r"""Wrap any function in a torch.nn.Module."""

from typing import Callable

from torch import Tensor
from torch.nn import Module


LambdaFunc = Callable[[Tensor], Tensor]


class LambdaLayer(Module):
    r"""Wrap any tensor operation in a network module."""

    def __init__(self, func: LambdaFunc) -> None:
        r"""Set callable tensor operation.

        Args:
            func: Callable tensor operation. Must be instance of ``torch.nn.Module``
                if it contains learnable parameters. In this case, however, the
                ``LambdaLayer`` wrapper becomes redundant. Main use is to wrap
                non-learnable Python functions.

        """
        super().__init__()
        self.func = func

    def forward(self, x: Tensor) -> Tensor:
        return self.func(x)



r"""Flow vector fields."""

from __future__ import annotations

from typing import Any, Optional, Sequence, Type, TypeVar, Union, overload

import torch
from torch import Tensor
from torch.nn import functional as F

from deepali.core import flow as U
from deepali.core.enum import PaddingMode, Sampling
from deepali.core.grid import ALIGN_CORNERS, Axes, Grid, grid_transform_vectors
from deepali.core.tensor import move_dim
from deepali.core.typing import Array, Device, DType, EllipsisType, PathStr, Scalar, ScalarOrTuple

from .image import Image, ImageBatch


TFlowField = TypeVar("TFlowField", bound="FlowField")
TFlowFields = TypeVar("TFlowFields", bound="FlowFields")


__all__ = ("FlowField", "FlowFields")


class FlowFields(ImageBatch):
    r"""Batch of flow fields."""

    def __init__(
        self: TFlowFields,
        data: Union[Array, ImageBatch],
        grid: Optional[Union[Grid, Sequence[Grid]]] = None,
        axes: Optional[Axes] = None,
        dtype: Optional[DType] = None,
        device: Optional[Device] = None,
        requires_grad: Optional[bool] = None,
        pin_memory: bool = False,
    ) -> None:
        r"""Initialize flow fields.

        Args:
            data: Batch data tensor of shape (N, D, ...X), where N is the batch size, and D
                must be equal the number of spatial dimensions. The order of the image channels
                must be such that vector components are in the order ``(x, ...)``.
            grid: Flow field sampling grids. If not otherwise specified, this attribute
                defines the fixed target image domain on which to resample a moving source image.
            axes: Axes with respect to which vectors are defined. By default, it is assumed that
                vectors are with respect to the unit ``grid`` cube in ``[-1, 1]^D``, where D are the
                number of spatial dimensions. If ``grid.align_corners() == False``, the extrema
                ``(-1, 1)`` refer to the boundary of the vector field ``grid``. Otherwise, the
                extrema coincide with the corner points of the sampling grid.
            dtype: Data type of the image data. A copy of the data is only made when the desired ``dtype``
                is not ``None`` and not the same as ``data.dtype``.
            device: Device on which to store image data. A copy of the data is only made when the data
                has to be copied to a different device.
            requires_grad: If autograd should record operations on the returned image tensor.
            pin_memory: If set, returned image tensor would be allocated in the pinned memory.
                Works only for CPU tensors.

        """
        # DataTensor.__new__() creates the tensor subclass given arguments:
        # data, dtype, device, requires_grad, pin_memory
        if grid is None and isinstance(data, ImageBatch):
            grid = data.grid()
            data = data.tensor()
        super().__init__(
            data,
            grid,
            dtype=dtype,
            device=device,
            requires_grad=requires_grad,
            pin_memory=pin_memory,
        )
        if self.shape[1] != self.sdim:
            raise ValueError(
                f"{type(self).__name__}() 'data' nchannels={self.shape[1]} must be equal spatial ndim={self.sdim}"
            )
        if axes is None:
            axes = Axes.from_grid(self._grid[0])
        else:
            axes = Axes.from_arg(axes)
        self._axes = axes

    def _make_instance(
        self: TFlowFields,
        data: Tensor,
        grid: Optional[Sequence[Grid]] = None,
        axes: Optional[Axes] = None,
        **kwargs,
    ) -> Union[TFlowFields, ImageBatch]:
        r"""Create a new instance while preserving subclass meta-data."""
        if data.shape[1] != data.ndim - 2:
            return ImageBatch(data, grid)
        kwargs["axes"] = axes or self._axes
        return super()._make_instance(data, grid, **kwargs)

    def _make_subitem(self, data: Tensor, grid: Grid) -> Union[FlowField, Image]:
        r"""Create FlowField in __getitem__. Can be overridden by subclasses to return a subtype."""
        if data.shape[0] == data.ndim - 1:
            return FlowField(data, grid, self._axes)
        return super()._make_subitem(data, grid)

    @staticmethod
    def _torch_function_axes(args) -> Optional[Axes]:
        r"""Get flow field Axes from args passed to __torch_function__."""
        if not args:
            return None
        if isinstance(args[0], (tuple, list)):
            args = args[0]
        axes: Sequence[Axes]
        axes = [ax for ax in (getattr(arg, "_axes", None) for arg in args) if ax is not None]
        if not axes:
            return None
        if any(ax != axes[0] for ax in axes[1:]):
            raise ValueError("Cannot apply __torch_function__ to flow fields with mismatching axes")
        return axes[0]

    @classmethod
    def _torch_function_result(
        cls, func, data, grid: Optional[Sequence[Grid]], axes: Optional[Axes]
    ) -> Any:
        if not isinstance(data, Tensor):
            return data
        if (
            grid
            and axes is not None
            and data.ndim == grid[0].ndim + 2
            and data.shape[1] == grid[0].ndim
            and data.shape[2:] == grid[0].shape
        ) or (grid is not None and not grid and data.ndim >= 4 and data.shape[0] == 0):
            if func in (torch.clone, Tensor.clone):
                grid = [g.clone() for g in grid]
            if isinstance(data, cls):
                data._grid = grid
                data._axes = axes
            else:
                data = cls(data, grid, axes)
        else:
            data = ImageBatch._torch_function_result(func, data, grid)
        return data

    @classmethod
    def __torch_function__(cls, func, types, args=(), kwargs=None):
        if func == F.grid_sample:
            raise ValueError("Argument of F.grid_sample() must be a batch, not a single image")
        if kwargs is None:
            kwargs = {}
        data = Tensor.__torch_function__(func, (Tensor,), args, kwargs)
        grid = cls._torch_function_grid(func, args, kwargs)
        axes = cls._torch_function_axes(args)
        if func in (
            torch.split,
            Tensor.split,
            torch.split_with_sizes,
            Tensor.split_with_sizes,
            torch.tensor_split,
            Tensor.tensor_split,
        ):
            return tuple(cls._torch_function_result(func, res, grid, axes) for res in data)
        return cls._torch_function_result(func, data, grid, axes)

    @overload
    def __getitem__(self: TFlowFields, index: int) -> FlowField:
        ...

    @overload
    def __getitem__(self: TFlowFields, index: EllipsisType) -> TFlowFields:
        ...

    @overload
    def __getitem__(self: TFlowFields, index: Union[list, slice, Tensor]) -> TFlowFields:
        ...

    def __getitem__(
        self: TFlowFields,
        index: Union[EllipsisType, int, slice, Sequence[Union[EllipsisType, int, slice]]],
    ) -> Union[FlowField, Image, TFlowFields, ImageBatch, Tensor]:
        r"""Get flow field at specified batch index, get a sub-batch, or extract region of interest tensor."""
        return super().__getitem__(index)

    @overload
    def axes(self: TFlowFields) -> Axes:
        r"""Get axes with respect to which flow vectors are defined."""
        ...

    @overload
    def axes(self: TFlowFields, axes: Axes) -> TFlowFields:
        r"""Get new batch of flow fields with flow vectors defined with respect to specified axes."""
        ...

    def axes(self: TFlowFields, axes: Optional[Axes] = None) -> Union[Axes, TFlowFields]:
        r"""Rescale and reorient vectors."""
        if axes is None:
            return self._axes
        data = self.tensor()
        data = move_dim(data, 1, -1)
        data = tuple(
            grid.transform_vectors(data[i : i + 1], axes=self._axes, to_axes=axes)
            for i, grid in enumerate(self._grid)
        )
        data = torch.cat(data, dim=0)
        data = move_dim(data, -1, 1)
        return self._make_instance(data, self._grid, axes)

    def curl(
        self: TFlowFields,
        mode: Optional[str] = None,
        sigma: Optional[float] = None,
        spacing: Optional[Union[Scalar, Array]] = None,
        stride: Optional[ScalarOrTuple[int]] = None,
    ) -> ImageBatch:
        if self.ndim not in (2, 3):
            raise RuntimeError(f"Cannot compute curl of {self.ndim}-dimensional flow field")
        if spacing is None:
            if self.axes() is Axes.GRID:
                spacing = 1
            elif self.axes() is Axes.WORLD:
                spacing = self.spacing()
            elif self.axes() is Axes.CUBE:
                spacing = tuple(2 / n for n in self.grid().size())
            else:
                spacing = tuple(2 / (n - 1) for n in self.grid().size())
        data = U.curl(self.tensor(), mode=mode, sigma=sigma, spacing=spacing, stride=stride)
        return ImageBatch(data, self._grid)

    def exp(
        self: TFlowFields,
        scale: Optional[float] = None,
        steps: Optional[int] = None,
        sampling: Union[Sampling, str] = Sampling.LINEAR,
        padding: Union[PaddingMode, str] = PaddingMode.BORDER,
    ) -> TFlowFields:
        r"""Group exponential maps of flow fields computed using scaling and squaring."""
        axes = self._axes
        align_corners = axes is Axes.CUBE_CORNERS
        flow = self.axes(Axes.CUBE_CORNERS if align_corners else Axes.CUBE)
        data = self.tensor()
        data = U.expv(
            data,
            scale=scale,
            steps=steps,
            sampling=sampling,
            padding=padding,
            align_corners=align_corners,
        )
        flow = self._make_instance(data, flow._grid, flow._axes)
        flow = flow.axes(axes)  # restore original axes
        return flow

    def sample(
        self: TFlowFields,
        arg: Union[Grid, Sequence[Grid], Tensor],
        mode: Optional[Union[Sampling, str]] = None,
        padding: Optional[Union[PaddingMode, str, Scalar]] = None,
    ) -> Union[TFlowFields, Tensor]:
        r"""Sample flow fields at optionally deformed unit grid points.

        Args:
            arg: Either a single grid which defines the sampling points for all images in the batch,
                a different grid for each image in the batch, or a tensor of normalized coordinates
                with shape ``(N, ..., D)`` or ``(1, ..., D)``. In the latter case, note that the
                shape ``...`` need not correspond to a (deformed) grid as required by ``grid_sample()``.
                It can be an arbitrary shape, e.g., ``M`` to sample at ``M`` given points.
            mode: Image interpolation mode.
            padding: Image extrapolation mode or scalar padding value.

        Returns:
            If ``arg`` is of type ``Grid`` or ``Sequence[Grid]``, a ``FlowFields`` batch is returned.
            When these grids match the grids of this batch of flow fields, ``self`` is returned.
            Otherwise, a ``Tensor`` of shape (N, C, ...) of sampled flow values is returned.
            Note that when ``arg`` is of type ``Grid`` or ``Sequence[Grid]``, flow vectors that are
            not expressed with respect to the world coordinate system will be implicitly converted to
            flow vectors with respect to the new sampling grids. If this is not desired, use a ``Tensor``
            type with sampling coordinates instead of ``Grid`` instances.

        """
        flow = super().sample(arg, mode=mode, padding=padding)
        if isinstance(flow, FlowFields):
            axes = flow.axes()
            if axes != Axes.WORLD:
                data = flow.tensor()
                data = U.move_dim(data, 1, -1)
                data = torch.cat(
                    [
                        grid_transform_vectors(v, grid, axes, to_grid, axes).unsqueeze_(0)
                        for v, grid, to_grid in zip(data, self._grid, flow.grids())
                    ],
                    dim=0,
                )
                data = U.move_dim(data, -1, 1)
                flow = self._make_instance(data, flow.grids())
        return flow

    def warp_image(
        self: TFlowFields,
        image: Union[Image, ImageBatch],
        sampling: Optional[Union[Sampling, str]] = None,
        padding: Optional[Union[PaddingMode, str]] = None,
    ) -> ImageBatch:
        r"""Deform given image (batch) using this batch of vector fields.

        Args:
            image: Single image or image batch. If a single ``Image`` is given, it is deformed by
                all the displacement fields in this batch. If an ``ImageBatch`` is given, the number
                of images in the batch must match the number of displacement fields in this batch.
            sampling: Interpolation mode for sampling values from ``image`` at deformed grid points.
            padding: Extrapolation mode for sampling values outside ``image`` domain.

        Returns:
            Batch of input images deformed by the vector fields of this batch.

        """
        if isinstance(image, Image):
            image = image.batch()
        align_corners = self._axes is Axes.CUBE_CORNERS
        grid = (g.coords(align_corners=align_corners, device=self.device) for g in self._grid)
        grid = torch.cat(tuple(g.unsqueeze(0) for g in grid), dim=0)
        flow = self.axes(Axes.from_align_corners(align_corners))
        flow = flow.tensor()
        flow = move_dim(flow, 1, -1)
        data = image.tensor()
        data = U.warp_image(
            data,
            grid,
            flow=flow,
            mode=sampling,
            padding=padding,
            align_corners=align_corners,
        )
        return image._make_instance(data, self._grid)

    def __repr__(self) -> str:
        return (
            type(self).__name__
            + f"(data={self.tensor()!r}, grids={self.grids()!r}, axes={self.axes()!r})"
        )

    def __str__(self) -> str:
        return (
            type(self).__name__
            + f"(data={self.tensor()!s}, grids={self.grids()!s}, axes={self.axes()!r})"
        )


class FlowField(Image):
    r"""Flow field image.

    A (dense) flow field is a vector image where the number of channels equals the number of spatial dimensions.
    The starting points of the vectors are defined on a regular oriented sampling grid positioned in world space.
    Orientation and scale of the vectors are defined with respect to a specified regular grid domain, which either
    coincides with the sampling grid, the world coordinate system, or the unit cube with side length 2 centered at
    the center of the sampling grid with axes parallel to the sampling grid. This unit cube domain is used by the
    ``torch.nn.functional.grid_sample()`` and ``torch.nn.functional.interpolate()`` functions.

    When a flow field is convert to a ``SimpleITK.Image``, the vectors are by default reoriented and rescaled such
    that these are with respect to the world coordinate system, a format common to ITK functions and other toolkits.

    """

    def __init__(
        self: TFlowField,
        data: Array,
        grid: Optional[Grid] = None,
        axes: Optional[Axes] = None,
        dtype: Optional[DType] = None,
        device: Optional[Device] = None,
        requires_grad: Optional[bool] = None,
        pin_memory: bool = False,
    ) -> None:
        r"""Initialize flow field.

        Args:
            data: Flow field data tensor of shape (C, ...X), where C must be equal the number of spatial dimensions.
                The order of the image channels must be such that vector components are in the order X, Y,...
            grid: Flow field sampling grid. If not otherwise specified, this attribute often also defines the fixed
                target image domain on which to resample a moving source image.
            axes: Axes with respect to which vectors are defined. By default, it is assumed that vectors are with
                respect to the unit ``grid`` cube in ``[-1, 1]^D``, where D are the number of spatial dimensions.
                If ``None`` and ``grid.align_corners() == False``, the extrema ``(-1, 1)`` refer to the boundary of
                the vector field ``grid``, and coincide with the grid corner points otherwise.
            dtype: Data type of the image data. A copy of the data is only made when the desired ``dtype``
                is not ``None`` and not the same as ``data.dtype``.
            device: Device on which to store image data. A copy of the data is only made when the data
                has to be copied to a different device.
            requires_grad: If autograd should record operations on the returned image tensor.
            pin_memory: If set, returned image tensor would be allocated in the pinned memory.
                Works only for CPU tensors.

        """
        # DataTensor.__new__() creates the tensor subclass given arguments:
        # data, dtype, device, requires_grad, pin_memory
        super().__init__(
            data,
            grid,
            dtype=dtype,
            device=device,
            requires_grad=requires_grad,
            pin_memory=pin_memory,
        )
        if self.nchannels != self._grid.ndim:
            raise ValueError(
                f"{type(self).__name__} nchannels={self.nchannels} must be equal grid.ndim={self._grid.ndim}"
            )
        if axes is None:
            axes = Axes.from_grid(self._grid)
        else:
            axes = Axes.from_arg(axes)
        self._axes = axes

    def _make_instance(
        self: TFlowField,
        data: Tensor,
        grid: Optional[Grid] = None,
        axes: Optional[Axes] = None,
        **kwargs,
    ) -> TFlowField:
        r"""Create a new instance while preserving subclass meta-data."""
        kwargs["axes"] = axes or self._axes
        return super()._make_instance(data, grid, **kwargs)

    @staticmethod
    def _torch_function_axes(args) -> Optional[Axes]:
        r"""Get flow field Axes from args passed to __torch_function__."""
        if not args:
            return None
        if isinstance(args[0], (tuple, list)):
            args = args[0]
        axes: Sequence[Axes]
        axes = [ax for ax in (getattr(arg, "_axes", None) for arg in args) if ax is not None]
        if not axes:
            return None
        if any(ax != axes[0] for ax in axes[1:]):
            raise ValueError("Cannot apply __torch_function__ to flow fields with mismatching axes")
        return axes[0]

    @classmethod
    def _torch_function_result(cls, func, data, grid: Optional[Grid], axes: Optional[Axes]) -> Any:
        if not isinstance(data, Tensor):
            return data
        if (
            grid is not None
            and axes is not None
            and data.ndim == grid.ndim + 1
            and data.shape[0] == grid.ndim
            and data.shape[1:] == grid.shape
        ):
            if func in (torch.clone, Tensor.clone):
                grid = grid.clone()
            if isinstance(data, cls):
                data._grid = grid
                data._axes = axes
            else:
                data = cls(data, grid, axes)
        else:
            data = Image._torch_function_result(func, data, grid)
        return data

    @classmethod
    def __torch_function__(cls, func, types, args=(), kwargs=None):
        if func == F.grid_sample:
            raise ValueError("Argument of F.grid_sample() must be a batch, not a single image")
        if kwargs is None:
            kwargs = {}
        data = Tensor.__torch_function__(func, (Tensor,), args, kwargs)
        grid = cls._torch_function_grid(args)
        axes = cls._torch_function_axes(args)
        if func in (
            torch.split,
            Tensor.split,
            torch.split_with_sizes,
            Tensor.split_with_sizes,
            torch.tensor_split,
            Tensor.tensor_split,
        ):
            return tuple(cls._torch_function_result(func, res, grid, axes) for res in data)
        return cls._torch_function_result(func, data, grid, axes)

    @classmethod
    def from_image(cls: Type[TFlowField], image: Image, axes: Optional[Axes] = None) -> TFlowField:
        r"""Create flow field from image instance."""
        return cls(image, image._grid, axes)

    def batch(self: TFlowField) -> FlowFields:
        r"""Batch of flow fields containing only this flow field."""
        data = self.unsqueeze(0)
        return FlowFields(data, self._grid, self._axes)

    @overload
    def axes(self: TFlowField) -> Axes:
        r"""Get axes with respect to which flow vectors are defined."""
        ...

    @overload
    def axes(self: TFlowField, axes: Axes) -> TFlowField:
        r"""Get new flow field with flow vectors defined with respect to specified axes."""
        ...

    def axes(self: TFlowField, axes: Optional[Axes] = None) -> Union[Axes, TFlowField]:
        r"""Rescale and reorient vectors with respect to specified axes."""
        if axes is None:
            return self._axes
        batch = self.batch()
        batch = batch.axes(axes)
        return batch[0]

    @classmethod
    def from_sitk(
        cls: Type[TFlowField],
        image: "sitk.Image",
        axes: Optional[Axes] = None,
        align_corners: bool = ALIGN_CORNERS,
        dtype: Optional[DType] = None,
        device: Optional[Device] = None,
    ) -> TFlowField:
        r"""Create vector field from ``SimpleITK.Image``."""
        image = Image.from_sitk(image, align_corners=align_corners, dtype=dtype, device=device)
        return cls.from_image(image, axes=axes or Axes.WORLD)

    def sitk(self: TFlowField, axes: Optional[Axes] = None) -> "sitk.Image":
        r"""Create ``SimpleITK.Image`` from this vector field."""
        disp: FlowField = self.detach()
        disp = disp.axes(axes or Axes.WORLD)
        return Image.sitk(disp)

    @classmethod
    def read(
        cls: Type[TFlowField],
        path: PathStr,
        axes: Optional[Axes] = None,
        align_corners: bool = ALIGN_CORNERS,
        dtype: Optional[DType] = None,
        device: Optional[Device] = None,
    ) -> TFlowField:
        r"""Read image data from file."""
        image = Image.read(path, align_corners=align_corners, dtype=dtype, device=device)
        return cls.from_image(image, axes=axes or Axes.WORLD)

    def write(self, path: PathStr, axes: Optional[Axes] = None, compress: bool = True) -> None:
        r"""Write flow field data to file."""
        disp: FlowField = self.detach()
        disp = disp.axes(axes or Axes.WORLD)
        Image.write(disp, path, compress=compress)

    def curl(
        self: TFlowField,
        mode: Optional[str] = None,
        sigma: Optional[float] = None,
        spacing: Optional[Union[Scalar, Array]] = None,
        stride: Optional[ScalarOrTuple[int]] = None,
    ) -> Image:
        r"""Compute curl of vector field."""
        batch = self.batch()
        rotvec = batch.curl(mode=mode, sigma=sigma, spacing=spacing, stride=stride)
        return rotvec[0]

    def exp(
        self: TFlowField,
        scale: Optional[float] = None,
        steps: Optional[int] = None,
        sampling: Union[Sampling, str] = Sampling.LINEAR,
        padding: Union[PaddingMode, str] = PaddingMode.BORDER,
    ) -> TFlowField:
        r"""Group exponential map of vector field computed using scaling and squaring."""
        batch = self.batch()
        batch = batch.exp(scale=scale, steps=steps, sampling=sampling, padding=padding)
        return batch[0]

    @overload
    def warp_image(
        self: TFlowField,
        image: Image,
        sampling: Optional[Union[Sampling, str]] = None,
        padding: Optional[Union[PaddingMode, str]] = None,
    ) -> Image:
        r"""Deform given image using this displacement field."""
        ...

    @overload
    def warp_image(
        self: TFlowField,
        image: ImageBatch,
        sampling: Optional[Union[Sampling, str]] = None,
        padding: Optional[Union[PaddingMode, str]] = None,
    ) -> ImageBatch:
        r"""Deform images in batch using this displacement field."""
        ...

    def warp_image(
        self: TFlowField,
        image: Union[Image, ImageBatch],
        sampling: Optional[Union[Sampling, str]] = None,
        padding: Optional[Union[PaddingMode, str]] = None,
    ) -> Union[Image, ImageBatch]:
        r"""Deform given image (batch) using this displacement field.

        Args:
            image: Single image or batch of images.
            kwargs: Keyword arguments to pass on to ``ImageBatch.warp()``.

        Returns:
            If ``image`` is an ``ImageBatch``, each image in the batch is deformed by this flow field
            and a batch of deformed images is returned. Otherwise, a single deformed image is returned.

        """
        batch = self.batch()
        result = batch.warp_image(image, sampling=sampling, padding=padding)
        if isinstance(image, Image) and len(result) == 1:
            return result[0]
        return result

    def __repr__(self) -> str:
        return (
            type(self).__name__
            + f"(data={self.tensor()!r}, grid={self.grid()!r}, axes={self.axes()!r})"
        )

    def __str__(self) -> str:
        return (
            type(self).__name__
            + f"(data={self.tensor()!s}, grid={self.grid()!s}, axes={self.axes()!r})"
        )



r"""Auxiliary functions for random sampling."""

import random
from typing import Optional, Union

from pkg_resources import parse_version

import torch
from torch import Generator, LongTensor, Tensor


def manual_seed(seed: int):
    r"""Seed all pseudo-random number generators."""
    random.seed(seed)
    try:
        import torch

        torch.manual_seed(seed)
    except ImportError:
        ...
    try:
        import numpy as np

        np.random.seed(seed)
    except ImportError:
        ...


def str_to_seed(seed: Union[int, str]) -> int:
    r"""Convert string argument to integer seed.

    Args:
        seed: Can be int, string of hexadecimal number, or other string.
            A hexadecimal number can optionally contain dashes. For example,
            can be data UUID (process ID) or MD5 image hash. When the input
            is neither an int nor hexadecimal number, the builtin ``hash``
            function is used to derive a seed from the given string.

    Returns:
        Seed value for random number generator.

    """
    if isinstance(seed, int):
        return seed
    if not isinstance(seed, str):
        raise TypeError("str_to_seed() 'seed' must be int or hexadecimal string")
    try:
        seed = int(seed, 0)
    except ValueError:
        try:
            seed = int(seed.replace("-", ""), 16)
        except ValueError:
            seed = abs(hash(seed))
    return seed


def multinomial(
    input: Tensor,
    num_samples: int,
    replacement: bool = False,
    generator: Optional[Generator] = None,
    out: Optional[LongTensor] = None,
) -> LongTensor:
    r"""Sample from a multinomial probability distribution.

    Args:
        input: Input vector of shape ``(N,)`` or matrix ``(M, N)``.
        num_samples: Number of random samples to draw.
        replacement: Whether to sample with or without replacement.
        generator: Random number generator to use.
        out: Pre-allocated output tensor.

    Returns:
        Indices of random samples. When ``input`` is a vector, a vector of ``num_samples`` indices
        is returned. Otherwise, a matrix of shape ``(M, num_samples)`` is returned. When ``out``
        is given, the returned tensor is a reference to ``out``.

    """
    if input.ndim == 0 or input.ndim > 2:
        raise ValueError("multinomial() 'input' must be vector or matrix")
    num_candidates = input.size(-1)
    if not replacement and num_candidates < num_samples:
        raise ValueError("multinomial() 'num_samples' cannot be greater than number of categories")
    if num_candidates > 2**24:
        impl = _multinomial
    else:
        impl = torch.multinomial
        input = input.float()
    return impl(input, num_samples, replacement=replacement, generator=generator, out=out)


def _multinomial(
    input: Tensor,
    num_samples: int,
    replacement: bool = False,
    generator: Optional[Generator] = None,
    out: Optional[LongTensor] = None,
) -> LongTensor:
    r"""Sample from a multinomial probability distribution.

    This function can be used for inputs of any size and is unlike ``torch.multinomial`` not limited
    to 2**24 categories at the expense of a less efficient implementation.

    Args:
        input: Input vector of shape ``(N,)`` or matrix ``(M, N)``.
        num_samples: Number of random samples to draw.
        replacement: Whether to sample with or without replacement.
        generator: Random number generator to use.
        out: Pre-allocated output tensor.

    Returns:
        Indices of random samples. When ``input`` is a vector, a vector of ``num_samples`` indices
        is returned. Otherwise, a matrix of shape ``(M, num_samples)`` is returned. When ``out``
        is given, the returned tensor is a reference to ``out``.

    """
    if input.ndim == 0 or input.ndim > 2:
        raise ValueError("_multinomial() 'input' must be vector or matrix")
    num_candidates = input.size(-1)
    out_shape = input.shape[:-1] + (num_samples,)
    if out is not None:
        if not isinstance(out, Tensor):
            raise TypeError("_multinomial() 'out' must be Tensor")
        if out.dtype != torch.int64:
            raise TypeError("_multinomial() 'out' must be int64 tensor")
        if out.shape != out_shape:
            raise ValueError(f"_multinomial() 'out' must have shape {out_shape}")
    # Use inverse transform sampling if the number of candidates is large and replacement=True
    if replacement:
        cdf = input.type(torch.float64).cumsum(dim=-1)
        cdf = cdf.div_(cdf[..., -1:].clone())
        val = torch.rand(out_shape, generator=generator, dtype=cdf.dtype, device=cdf.device)
        out = torch.searchsorted(cdf, val, out=out).clip_(0, num_candidates - 1)
    # In case of replacement=False, use Gumbel-max trick instead of inverse transform sampling.
    else:
        if num_samples > num_candidates:
            raise ValueError(
                "_multinomial() 'num_samples' cannot be greater than number of categories"
            )
        logit = input.log()
        value = torch.rand(
            input.shape[:-1] + (num_candidates,),
            generator=generator,
            dtype=logit.dtype,
            device=input.device,
        )
        value = value.log_().neg_().log_().neg_().add_(logit)
        if parse_version(torch.__version__) < parse_version("1.12"):
            _, index = torch.topk(value, num_samples, dim=-1, sorted=False)
            out = index if out is None else out.copy_(index)
        else:
            if out is None:
                out = torch.empty(out_shape, dtype=torch.int64, device=value.device)
            _ = torch.empty(out_shape, dtype=value.dtype, device=value.device)
            _, out = torch.topk(value, num_samples, dim=-1, sorted=False, out=(_, out))
    return out



from typing import Optional, Sequence

import numpy as np
import SimpleITK as sitk


def image_dtype(image: sitk.Image) -> np.dtype:
    r"""Get NumPy data type of SimpleITK image."""
    return sitk.GetArrayViewFromImage(image).dtype


def image_from_array(
    values: np.ndarray,
    origin: Optional[Sequence[float]] = None,
    spacing: Optional[Sequence[float]] = None,
    direction: Optional[Sequence[float]] = None,
) -> sitk.Image:
    r"""Create image from sampling grid and image data array."""
    components = 1
    ndim = 0
    if origin:
        ndim = len(origin)
    elif spacing:
        ndim = len(spacing)
    elif direction:
        ndim = int(np.sqrt(len(direction)))
        assert len(direction) == ndim * ndim
    if ndim:
        if values.ndim < ndim:
            values = values.reshape((1,) * (ndim - values.ndim) + values.shape)
        elif values.ndim > ndim:
            values = values.reshape(values.shape[:ndim] + (-1,))
            components = values.shape[-1]
        assert ndim <= values.ndim <= ndim + 1
    image = sitk.GetImageFromArray(values, isVector=components > 1)
    if origin:
        image.SetOrigin(origin)
    if spacing:
        image.SetSpacing(spacing)
    if direction:
        image.SetDirection(direction)
    return image


def array_from_image(image: sitk.Image, view: bool = False) -> np.ndarray:
    r"""Get NumPy array with copy of image values."""
    return sitk.GetArrayViewFromImage(image) if view else sitk.GetArrayFromImage(image)



r"""Image data transforms."""

from pathlib import Path
from typing import List, Mapping, Optional, Sequence, Union

import torch
from torch import Tensor
from torch.nn import Module

from deepali.core.enum import PaddingMode, Sampling
from deepali.core.typing import DeviceStr, DTypeStr, PathUri, ScalarOrTuple

from ..image import Image
from .item import ItemTransform, ItemwiseTransform


__all__ = (
    "AvgPoolImage",
    "CastImage",
    "CenterCropImage",
    "CenterPadImage",
    "ClampImage",
    "ImageToTensor",
    "NarrowImage",
    "NormalizeImage",
    "ReadImage",
    "ResampleImage",
    "RescaleImage",
    "ResizeImage",
    "ImageTransformConfig",
    "config_has_read_image_transform",
    "prepend_read_image_transform",
    "image_transform",
    "image_transforms",
)


class AvgPoolImage(ItemwiseTransform, Module):
    r"""Downsample image using average pooling."""

    def __init__(
        self,
        kernel_size: ScalarOrTuple[int],
        stride: Optional[ScalarOrTuple[int]] = None,
        padding: ScalarOrTuple[int] = 0,
        ceil_mode: bool = False,
        count_include_pad: bool = True,
    ) -> None:
        super().__init__()
        self.kernel_size = kernel_size
        self.stride = stride
        self.padding = padding
        self.ceil_mode = ceil_mode
        self.count_include_pad = count_include_pad

    def forward(self, image: Image) -> Image:
        if not isinstance(image, Image):
            raise TypeError(f"{type(self).__name__}.forward() argument must be Image")
        return image.avg_pool(
            self.kernel_size,
            stride=self.stride,
            padding=self.padding,
            ceil_mode=self.ceil_mode,
            count_include_pad=self.count_include_pad,
        )

    def __repr__(self) -> str:
        return (
            type(self).__name__
            + f"(kernel_size={self.kernel_size!r},"
            + f" stride={self.stride!r},"
            + f" padding={self.padding!r},"
            + f" ceil_mode={self.ceil_mode},"
            + f" count_include_pad={self.count_include_pad})"
        )


class CastImage(ItemwiseTransform, Module):
    r"""Cast image data to specified type."""

    def __init__(self, dtype: DTypeStr) -> None:
        super().__init__()
        if isinstance(dtype, str):
            attr = dtype
            dtype = getattr(torch, attr, None)
            if dtype is None:
                raise ValueError(
                    f"{type(self).__name__}() module torch has no 'dtype' named torch.{attr}"
                )
            if not isinstance(dtype, torch.dtype):
                raise ValueError(f"{type(self).__name__}() torch.{attr} is not a torch.dtype name")
        elif not isinstance(dtype, torch.dtype):
            raise TypeError(f"{type(self).__name__}() 'dtype' must by dtype name or torch.dtype")
        self.dtype = dtype

    def forward(self, image: Image) -> Image:
        if not isinstance(image, Image):
            raise TypeError(f"{type(self).__name__}.forward() argument must be Image")
        return image.type(self.dtype)

    def __repr__(self) -> str:
        return type(self).__name__ + f"(dtype={self.dtype!r})"


class CenterCropImage(ItemwiseTransform, Module):
    r"""Crop image to specified maximum output size."""

    def __init__(self, size: Union[int, Sequence[int]]) -> None:
        super().__init__()
        self.size = size

    def forward(self, image: Image) -> Image:
        if not isinstance(image, Image):
            raise TypeError(f"{type(self).__name__}.forward() argument must be Image")
        return image.center_crop(self.size)

    def __repr__(self) -> str:
        return type(self).__name__ + f"(size={self.size!r})"


class CenterPadImage(ItemwiseTransform, Module):
    r"""Pad image to specified minimum output size."""

    def __init__(
        self,
        size: Union[int, Sequence[int]],
        mode: Union[PaddingMode, str] = PaddingMode.CONSTANT,
        value: float = 0,
    ) -> None:
        super().__init__()
        self.size = size
        self.mode = PaddingMode(mode)
        self.value = float(value)

    def forward(self, image: Image) -> Image:
        if not isinstance(image, Image):
            raise TypeError(f"{type(self).__name__}.forward() argument must be Image")
        return image.center_pad(self.size, mode=self.mode, value=self.value)

    def __repr__(self) -> str:
        return (
            type(self).__name__
            + f"(size={self.size!r}, mode={self.mode.value!r}, value={self.value!r})"
        )


class ClampImage(ItemwiseTransform, Module):
    r"""Clamp image intensities to specified minimum and/or maximum value."""

    def __init__(
        self,
        min: Optional[float] = None,
        max: Optional[float] = None,
        inplace: bool = False,
    ) -> None:
        super().__init__()
        self.min = min
        self.max = max
        self.inplace = bool(inplace)

    def forward(self, image: Image) -> Image:
        if not isinstance(image, Image):
            raise TypeError(f"{type(self).__name__}.forward() argument must be Image")
        clamp_fn = image.clamp_ if self.inplace else image.clamp
        image = clamp_fn(self.min, self.max)
        return image

    def __repr__(self) -> str:
        return (
            type(self).__name__ + f"(min={self.min!r}, max={self.max!r}, inplace={self.inplace!r})"
        )


class ImageToTensor(ItemwiseTransform, Module):
    r"""Convert image to data tensor."""

    def forward(self, image: Image) -> Tensor:
        if not isinstance(image, Image):
            raise TypeError(f"{type(self).__name__}.forward() argument must be Image")
        return image.tensor()

    def __repr__(self) -> str:
        return type(self).__name__ + "()"


class NarrowImage(ItemwiseTransform, Module):
    r"""Return image with data tensor narrowed along specified dimension."""

    def __init__(self, dim: int, start: int, length: int = 1) -> None:
        super().__init__()
        if dim != 0:
            raise NotImplementedError(
                "NarrowImage() 'dim' must be zero at the moment."
                "Extend implementation to adjust image.grid() for other image dimensions."
            )
        self.dim = dim
        self.start = start
        self.length = length

    def forward(self, image: Image) -> Tensor:
        if not isinstance(image, Image):
            raise TypeError(f"{type(self).__name__}.forward() argument must be Image")
        image = image.narrow(self.dim, self.start, self.length)
        return image

    def __repr__(self) -> str:
        return type(self).__name__ + f"(dim={self.dim}, start={self.start}, length={self.length})"


class NormalizeImage(ItemwiseTransform, Module):
    r"""Normalize and clamp image intensities in [min, max]."""

    def __init__(
        self,
        min: Optional[float] = None,
        max: Optional[float] = None,
        mode: str = "unit",
        inplace: bool = False,
    ) -> None:
        super().__init__()
        if mode not in ("center", "unit", "zscore", "z-score"):
            raise ValueError("NormalizeImage() 'mode' must be 'center', 'unit', or 'zscore'")
        self.min = min
        self.max = max
        self.mode = mode
        self.inplace = inplace

    def forward(self, image: Image) -> Image:
        if not isinstance(image, Image):
            raise TypeError(f"{type(self).__name__}.forward() argument must be Image")
        normalize_fn = image.normalize_ if self.inplace else image.normalize
        return normalize_fn(mode=self.mode, min=self.min, max=self.max)

    def __repr__(self) -> str:
        return (
            type(self).__name__
            + f"(mode={self.mode!r}, min={self.min!r}, max={self.max!r}, inplace={self.inplace!r})"
        )


class ReadImage(ItemwiseTransform, Module):
    r"""Read image data from file path."""

    def __init__(
        self,
        dtype: Optional[DTypeStr] = None,
        device: Optional[DeviceStr] = None,
    ) -> None:
        super().__init__()
        if isinstance(dtype, str):
            attr = dtype
            dtype = getattr(torch, attr, None)
            if dtype is None:
                raise ValueError(f"ReadImage() module torch has no 'dtype' named torch.{attr}")
        if dtype is not None and not isinstance(dtype, torch.dtype):
            raise TypeError("ReadImage() 'dtype' must by None or torch.dtype")
        self.dtype = dtype
        self.device = torch.device(device or "cpu")

    def forward(self, path: PathUri) -> Image:
        if not isinstance(path, (str, Path)):
            raise TypeError(f"{type(self).__name__}() 'path' must be Path or str")
        image = Image.read(path, dtype=self.dtype, device=self.device)
        return image

    def __repr__(self) -> str:
        return type(self).__name__ + f"(dtype={self.dtype}, device='{self.device!s}')"


class ResampleImage(ItemwiseTransform, Module):
    r"""Resample image to specified voxel size."""

    def __init__(
        self,
        spacing: Union[float, Sequence[float], str],
        mode: Union[Sampling, str] = Sampling.LINEAR,
    ) -> None:
        super().__init__()
        self.spacing = spacing
        self.mode = Sampling(mode)

    def forward(self, image: Image) -> Image:
        if not isinstance(image, Image):
            raise TypeError(f"{type(self).__name__}.forward() argument must be Image")
        image = image.resample(self.spacing, mode=self.mode)
        return image

    def __repr__(self) -> str:
        return type(self).__name__ + f"(spacing={self.spacing!r}, mode={self.mode.value!r})"


class RescaleImage(ItemwiseTransform, Module):
    r"""Linearly rescale image data."""

    def __init__(
        self,
        min: Optional[float] = None,
        max: Optional[float] = None,
        mul: Optional[float] = None,
        add: Optional[float] = None,
    ) -> None:
        super().__init__()
        if mul is not None or add is not None:
            if min is not None or max is not None:
                raise ValueError(
                    "RescaleImage() 'min'/'max' and 'add'/'mul' are mutually exclusive"
                )
            self.min = None
            self.max = None
            self.mul = 1 if mul is None else float(mul)
            self.add = 0 if add is None else float(add)
        else:
            self.min = min
            self.max = max
            self.mul = None
            self.add = None

    def forward(self, image: Image) -> Image:
        if not isinstance(image, Image):
            raise TypeError(f"{type(self).__name__}.forward() argument must be Image")
        if self.mul is not None or self.add is not None:
            assert self.min is None and self.max is None
            if self.mul != 1:
                image = image.mul(self.mul)
            if self.add != 0:
                image = image.add(self.add)
        else:
            image = image.rescale(min=self.min, max=self.max)
        return image

    def __repr__(self) -> str:
        s = type(self).__name__ + "("
        if self.mul is not None or self.add is not None:
            s += f"mul={self.mul!r}, add={self.add!r}"
        else:
            s += f"min={self.min!r}, max={self.max!r}"
        return s


class ResizeImage(ItemwiseTransform, Module):
    r"""Resample image to specified image size."""

    def __init__(
        self,
        size: Union[int, Sequence[int]],
        mode: Union[Sampling, str] = Sampling.LINEAR,
    ) -> None:
        super().__init__()
        self.size = size
        self.mode = Sampling(mode)

    def forward(self, image: Image) -> Image:
        if not isinstance(image, Image):
            raise TypeError(f"{type(self).__name__}.forward() argument must be Image")
        image = image.resize(self.size, mode=self.mode)
        return image

    def __repr__(self) -> str:
        return type(self).__name__ + f"(size={self.size!r}, mode={self.mode.value!r})"


ImageTransformMapping = Mapping[str, Union[Sequence, Mapping]]
ImageTransformConfig = Union[
    str, ImageTransformMapping, Sequence[Union[str, ImageTransformMapping]]
]

IMAGE_TRANSFORM_TYPES = {
    "avgpool": AvgPoolImage,
    "cast": CastImage,
    "centercrop": CenterCropImage,
    "centerpad": CenterPadImage,
    "clamp": ClampImage,
    "narrow": NarrowImage,
    "normalize": NormalizeImage,
    "read": ReadImage,
    "rescale": RescaleImage,
    "resample": ResampleImage,
    "resize": ResizeImage,
}

INPLACE_IMAGE_TRANSFORMS = {"clamp"}


def config_has_read_image_transform(config: ImageTransformConfig) -> bool:
    r"""Whether image data transformation configuration contains a "read" image transform."""
    if isinstance(config, str):
        return config.lower() == "read"
    if isinstance(config, Mapping):
        for name in config:
            if name.lower() == "read":
                return True
        return False
    if isinstance(config, Sequence):
        for item in config:
            if isinstance(item, Sequence) and not isinstance(item, str):
                raise ValueError(
                    "config_has_read_image_transform() 'config' Sequence cannot be nested"
                )
            if config_has_read_image_transform(item):
                return True
        return False
    raise TypeError("config_has_read_image_transform() 'config' must be str, Mapping, or Sequence")


def prepend_read_image_transform(
    config: ImageTransformConfig,
    dtype: Optional[str] = None,
    device: Optional[str] = None,
) -> ImageTransformConfig:
    r"""Insert a "read" image transform before any other image data transform."""
    if config_has_read_image_transform(config):
        return config
    read_transform_config = {"read": dict(dtype=dtype, device=device)}
    if isinstance(config, str):
        return [read_transform_config, config]
    if isinstance(config, Mapping):
        return {**read_transform_config, **config}
    if isinstance(config, Sequence):
        return [read_transform_config] + list(config)
    raise TypeError("prepend_read_image_transform() 'config' must be str, Mapping, or Sequence")


def image_transform(
    name: str, *args, key: Optional[str] = None, inplace: bool = True, **kwargs
) -> ItemwiseTransform:
    r"""Create image data transform given its name."""
    cls = IMAGE_TRANSFORM_TYPES.get(name.lower())
    if cls is None:
        raise ValueError(f"image_transform() unknown image data transform '{name}'")
    if name in INPLACE_IMAGE_TRANSFORMS:
        kwargs["inplace"] = inplace
    transform = cls(*args, **kwargs)
    if key:
        transform = ItemTransform(transform, key=key)
    return transform


def image_transforms(config: ImageTransformConfig, key: Optional[str] = None) -> List[Module]:
    r"""Create image data transforms from configuration.

    A sequence of image transforms can be configured using names:

    - ``avgpool``: :class:`.AvgPoolImage`
    - ``cast``: :class:`.CastImage`
    - ``centercrop``: :class:`.CenterCropImage`
    - ``centerpad``: :class:`.CenterPadImage`
    - ``clamp``: :class:`.ClampImage`
    - ``narrow``: :class:`.NarrowImage`
    - ``normalize``: :class:`.NormalizeImage`
    - ``read``: :class:`.ReadImage`
    - ``rescale``: :class:`.RescaleImage`
    - ``resample``: :class:`.ResampleImage`
    - ``resize``: :class:`.ResizeImage`

    Parameters can be passed to the data transform as keyword arguments, e.g.,

    .. code-block:: yaml

        transforms:
        - read: {dtype: float32}
        - avgpool: {kernel_size: 2}
        - normalize: {min: 0, max: 255, mode: unit}

    """
    transforms = []
    if isinstance(config, str):
        transforms.append(image_transform(config, key=key))
    elif isinstance(config, Mapping):
        for name, value in config.items():
            if value is None:
                transform = image_transform(name, key=key)
            elif isinstance(value, (list, tuple)):
                transform = image_transform(name, *value, key=key)
            elif isinstance(value, Mapping):
                transform = image_transform(name, key=key, **value)
            else:
                transform = image_transform(name, value, key=key)
            transforms.append(transform)
    elif isinstance(config, Sequence):
        for item in config:
            if isinstance(item, Sequence) and not isinstance(item, str):
                raise ValueError("image_transform() 'config' Sequence cannot be nested")
            transforms.extend(image_transforms(item, key=key))
    else:
        raise TypeError("image_transforms() 'config' must be str, Mapping, or Sequence")
    return transforms



r"""Cubic B-spline free-form deformations."""

from __future__ import annotations

from copy import copy as shallow_copy
from typing import cast, overload
from typing import Callable, List, Optional, Sequence, Tuple, TypeVar, Union

import torch
from torch import Size, Tensor

from deepali.core import kernels as K
from deepali.core import functional as U
from deepali.core.enum import SpatialDim
from deepali.core.grid import Grid
from deepali.core.typing import ScalarOrTuple
from deepali.modules import ExpFlow

from .base import NonRigidTransform
from .parametric import ParametricTransform


TBSplineTransform = TypeVar("TBSplineTransform", bound="BSplineTransform")


class BSplineTransform(ParametricTransform, NonRigidTransform):
    r"""Non-rigid transformation parameterized by cubic B-spline function."""

    def __init__(
        self,
        grid: Grid,
        groups: Optional[int] = None,
        params: Optional[Union[bool, Tensor, Callable[..., Tensor]]] = True,
        stride: Optional[Union[int, Sequence[int]]] = None,
        transpose: bool = False,
    ) -> None:
        r"""Initialize transformation parameters.

        Args:
            grid: Grid domain on which transformation is defined.
            groups: Number of transformations. A given image batch can either be deformed by a
                single transformation, or a separate transformation for each image in the batch, e.g.,
                for group-wise or batched registration. The default is one transformation for all images
                in the batch, or the batch length of the ``params`` tensor if provided.
            params: Initial parameters. If a tensor is given, it is only registered as optimizable module
                parameters when of type ``torch.nn.Parameter``. When a callable is given instead, it will be
                called by ``self.update()`` with arguments set and given by ``self.condition()``. When a boolean
                argument is given, a new zero-initialized tensor is created. If ``True``, this tensor is registered
                as optimizable module parameter.
            stride: Number of grid points between control points plus one. This is the stride of the
                transposed convolution used to upsample the control point displacements to the sampling ``grid``
                size. If ``None``, a stride of 1 is used. If a sequence of values is given, these must be the
                strides for the different spatial grid dimensions in the order ``(sx, sy, sz)``. Note that
                when the control point grid is subdivided in order to double its size along each spatial
                dimension, the stride with respect to this subdivided control point grid remains the same.
            transpose: Whether to use separable transposed convolution as implemented in AIRLab.
                When ``False``, a more efficient implementation using multi-channel convolution followed
                by a reshuffling of the output is performed. This more efficient and also more accurate
                implementation is adapted from the C++ code of MIRTK (``mirtk::BSplineInterpolateImageFunction``).

        """
        if not grid.align_corners():
            raise ValueError("BSplineTransform() requires 'grid.align_corners() == True'")
        if stride is None:
            stride = 5
        if isinstance(stride, int):
            stride = (stride,) * grid.ndim
        if len(stride) != grid.ndim:
            raise ValueError(f"BSplineTransform() 'stride' must be single int or {grid.ndim} ints")
        self.stride = tuple(int(s) for s in stride)
        self._transpose = transpose  # MUST be set before register_kernels() is called
        super().__init__(grid, groups=groups, params=params)
        self.register_kernels(stride)

    @property
    def data_shape(self) -> Size:
        r"""Get shape of transformation parameters tensor."""
        grid = self.grid()
        shape = U.cubic_bspline_control_point_grid_size(grid.shape, self.data_stride)
        return Size((grid.ndim,) + shape)

    @property
    def data_stride(self) -> Tuple[int, ...]:
        return tuple(reversed([int(s) for s in self.stride]))

    @torch.no_grad()
    def grid_(self: TBSplineTransform, grid: Grid) -> TBSplineTransform:
        r"""Set sampling grid of transformation domain and codomain.

        If ``self.params`` is a callable, only the grid attribute is updated, and
        the callable must return a tensor of matching size upon next evaluation.

        Args:
            grid: New sampling grid for dense displacement field at which FFD is evaluated.
                This function currently only supports subdivision of the control point grid,
                i.e., the new ``grid`` must have size ``2 * n - 1`` along each spatial dimension
                that should be subdivided, where ``n`` is the current grid size, or have the same
                size as the current grid for dimensions that remain the same.

        Returns:
            Reference to this modified transformation object.

        """
        params = self.params
        current_grid = self._grid
        if grid.ndim != current_grid.ndim:
            raise ValueError(
                f"{type(self).__name__}.grid_() argument must have {current_grid.ndim} dimensions"
            )
        subdivide_dims: List[SpatialDim] = []
        if isinstance(params, Tensor):
            current_grid = self._grid
            if grid.ndim != current_grid.ndim:
                raise ValueError(
                    f"{type(self).__name__}.grid_() argument must have {current_grid.ndim} dimensions"
                )
            if not grid.align_corners():
                raise ValueError(
                    f"{type(self).__name__}() requires grid.align_corners() to be True"
                )
            if not grid.same_domain_as(current_grid):
                raise ValueError(
                    f"{type(self).__name__}.grid_() argument must define same grid domain as current grid"
                )
            new_size = grid.size()
            current_size = current_grid.size()
            for i in range(grid.ndim):
                if new_size[i] == 2 * current_size[i] - 1:
                    subdivide_dims.append(SpatialDim(i))
                elif new_size[i] != current_size[i]:
                    raise ValueError(
                        f"{type(self).__name__}.grid_() argument must have same size or new size '2n - 1'"
                    )
        self._grid = grid
        if subdivide_dims:
            new_shape = (params.shape[0],) + self.data_shape
            new_params = U.subdivide_cubic_bspline(params, dims=subdivide_dims)
            for dim in subdivide_dims:
                dim = dim.tensor_dim(params.ndim)
                new_params = new_params.narrow(dim, 1, new_shape[dim])
            self.data_(new_params.contiguous())
        return self

    @staticmethod
    def kernel_name(stride: int) -> str:
        r"""Get name of buffer for 1-dimensional kernel for given control point spacing."""
        return "kernel_stride_" + str(stride)

    @overload
    def kernel(self) -> Tuple[Tensor, ...]:
        ...

    @overload
    def kernel(self, stride: int) -> Tensor:
        ...

    @overload
    def kernel(self, stride: Sequence[int]) -> Tuple[Tensor, ...]:
        ...

    def kernel(
        self, stride: Optional[ScalarOrTuple[int]] = None
    ) -> Union[Tensor, Tuple[Tensor, ...]]:
        r"""Get 1-dimensional kernels for given control point spacing."""
        if stride is None:
            stride = self.stride
        if isinstance(stride, int):
            return getattr(self, self.kernel_name(stride))
        return tuple(getattr(self, self.kernel_name(s)) for s in stride)

    def register_kernels(self, stride: Union[int, Sequence[int]]) -> None:
        r"""Precompute cubic B-spline kernels."""
        if isinstance(stride, int):
            stride = [stride]
        for s in stride:
            name = self.kernel_name(s)
            if not hasattr(self, name):
                if self._transpose:
                    kernel = K.cubic_bspline1d(s)
                else:
                    kernel = U.bspline_interpolation_weights(degree=3, stride=s)
                self.register_buffer(name, kernel, persistent=False)

    def deregister_kernels(self, stride: Union[int, Sequence[int]]) -> None:
        r"""Remove precomputed cubic B-spline kernels."""
        if isinstance(stride, int):
            stride = [stride]
        for s in stride:
            name = self.kernel_name(s)
            if hasattr(self, name):
                delattr(self, name)

    def evaluate_spline(self) -> Tensor:
        r"""Evaluate cubic B-spline at sampling grid points."""
        data = self.data()
        grid = self.grid()
        if not grid.align_corners():
            raise AssertionError(
                f"{type(self).__name__}() requires grid.align_corners() to be True"
            )
        stride = self.stride
        kernel = self.kernel(stride)
        u = U.evaluate_cubic_bspline(
            data,
            shape=grid.shape,
            stride=stride,
            kernel=kernel,
            transpose=self._transpose,
        )
        return u


class FreeFormDeformation(BSplineTransform):
    r"""Cubic B-spline free-form deformation model."""

    def update(self) -> FreeFormDeformation:
        r"""Update buffered displacement vector field."""
        super().update()
        u = self.evaluate_spline()
        self.register_buffer("u", u, persistent=False)
        return self


class StationaryVelocityFreeFormDeformation(BSplineTransform):
    r"""Stationary velocity field based transformation model using cubic B-spline parameterization."""

    def __init__(
        self,
        grid: Grid,
        groups: Optional[int] = None,
        params: Optional[Union[bool, Tensor, Callable[..., Tensor]]] = True,
        stride: Optional[Union[int, Sequence[int]]] = None,
        scale: Optional[float] = None,
        steps: Optional[int] = None,
        transpose: bool = False,
    ) -> None:
        r"""Initialize transformation parameters.

        Args:
            grid: Grid on which to sample flow field vectors.
            groups: Number of velocity fields.
            params: Initial parameters of cubic B-spline velocity fields of shape ``(N, C, ...X)``.
            stride: Number of grid points between control points (minus one).
            scale: Constant scaling factor of velocity fields.
            steps: Number of scaling and squaring steps.
            transpose: Whether to use separable transposed convolution as implemented in AIRLab.
                When ``False``, a more efficient implementation using multi-channel convolution followed
                by a reshuffling of the output is performed. This more efficient and also more accurate
                implementation is adapted from the C++ code of MIRTK (``mirtk::BSplineInterpolateImageFunction``).

        """
        align_corners = grid.align_corners()
        super().__init__(grid, groups=groups, params=params, stride=stride, transpose=transpose)
        self.exp = ExpFlow(scale=scale, steps=steps, align_corners=align_corners)

    def inverse(
        self, link: bool = False, update_buffers: bool = False
    ) -> StationaryVelocityFreeFormDeformation:
        r"""Get inverse of this transformation.

        Args:
            link: Whether the inverse transformation keeps a reference to this transformation.
                If ``True``, the ``update()`` function of the inverse function will not recompute
                shared parameters (e.g., parameters obtained by a callable neural network), but
                directly access the parameters from this transformation. Note that when ``False``,
                the inverse transformation will still share parameters, modules, and buffers with
                this transformation, but these shared tensors may be replaced by a call of ``update()``
                (which is implicitly called as pre-forward hook when ``__call__()`` is invoked).
            update_buffers: Whether buffers of inverse transformation should be updated after creating
                the shallow copy. If ``False``, the ``update()`` function of the returned inverse
                transformation has to be called before it is used.

        Returns:
            Shallow copy of this transformation with ``exp`` module which uses negative scaling factor
            to scale and square the stationary velocity field to compute the inverse displacement field.

        """
        inv = shallow_copy(self)
        if link:
            inv.link_(self)
        inv.exp = cast(ExpFlow, self.exp).inverse()
        if update_buffers:
            v = getattr(inv, "v", None)
            if v is not None:
                u = inv.exp(v)
                inv.register_buffer("u", u, persistent=False)
        return inv

    def update(self) -> StationaryVelocityFreeFormDeformation:
        r"""Update buffered velocity and displacement vector fields."""
        super().update()
        v = self.evaluate_spline()
        u = self.exp(v)
        self.register_buffer("v", v, persistent=False)
        self.register_buffer("u", u, persistent=False)
        return self



r"""Image dissimilarity measures."""

from typing import Optional, Union

from torch import Tensor

from deepali.core import functional as U
from deepali.core.typing import ScalarOrTuple

from . import functional as L
from .base import NormalizedPairwiseImageLoss
from .base import PairwiseImageLoss


class Dice(PairwiseImageLoss):
    r"""Generalized Sorensen-Dice similarity coefficient."""

    def __init__(self, epsilon: float = 1e-15) -> None:
        super().__init__()
        self.epsilon = epsilon

    def forward(self, source: Tensor, target: Tensor, mask: Optional[Tensor] = None) -> Tensor:
        r"""Evaluate image dissimilarity loss."""
        return L.dice_loss(source, target, weight=mask, epsilon=self.epsilon)

    def extra_repr(self) -> str:
        return f"epsilon={self.epsilon:.2e}"


DSC = Dice


class NCC(PairwiseImageLoss):
    r"""Normalized cross correlation."""

    def __init__(self, epsilon: float = 1e-15) -> None:
        super().__init__()
        self.epsilon = epsilon

    def forward(self, source: Tensor, target: Tensor, mask: Optional[Tensor] = None) -> Tensor:
        r"""Evaluate image dissimilarity loss."""
        return L.ncc_loss(
            source,
            target,
            mask=mask,
            epsilon=self.epsilon,
        )

    def extra_repr(self) -> str:
        return f"epsilon={self.epsilon:.2e}"


class LCC(PairwiseImageLoss):
    r"""Local normalized cross correlation."""

    def __init__(self, kernel_size: ScalarOrTuple[int] = 7, epsilon: float = 1e-15) -> None:
        super().__init__()
        self.kernel_size = kernel_size
        self.epsilon = epsilon

    def forward(self, source: Tensor, target: Tensor, mask: Optional[Tensor] = None) -> Tensor:
        r"""Evaluate image dissimilarity loss."""
        return L.lcc_loss(
            source,
            target,
            mask=mask,
            kernel_size=self.kernel_size,
            epsilon=self.epsilon,
        )

    def extra_repr(self) -> str:
        return f"kernel_size={self.kernel_size}, epsilon={self.epsilon:.2e}"


LNCC = LCC


class WLCC(PairwiseImageLoss):
    r"""Weighted local normalized cross correlation."""

    def __init__(self, kernel_size: ScalarOrTuple[int] = 7, epsilon: float = 1e-15) -> None:
        super().__init__()
        self.kernel_size = kernel_size
        self.epsilon = epsilon

    def forward(
        self,
        source: Tensor,
        target: Tensor,
        mask: Optional[Tensor] = None,
        source_mask: Optional[Tensor] = None,
        target_mask: Optional[Tensor] = None,
    ) -> Tensor:
        r"""Evaluate image dissimilarity loss."""
        return L.wlcc_loss(
            source,
            target,
            mask=mask,
            source_mask=source_mask,
            target_mask=target_mask,
            kernel_size=self.kernel_size,
            epsilon=self.epsilon,
        )

    def extra_repr(self) -> str:
        return f"kernel_size={self.kernel_size}, epsilon={self.epsilon:.2e}"


SLCC = WLCC  # Sparse local cross correlation (cf. Lewis et al. 2020, arXiv:1812.06932)


class L1ImageLoss(NormalizedPairwiseImageLoss):
    r"""Average absolute intensity differences."""

    def forward(self, source: Tensor, target: Tensor, mask: Optional[Tensor] = None) -> Tensor:
        r"""Evaluate image dissimilarity loss."""
        return L.mae_loss(source, target, mask=mask, norm=self.norm)


MAE = L1ImageLoss


class HuberImageLoss(NormalizedPairwiseImageLoss):
    r"""Average Huber loss."""

    def __init__(
        self,
        source: Optional[Tensor] = None,
        target: Optional[Tensor] = None,
        norm: Optional[Union[bool, Tensor]] = None,
        delta: Optional[float] = None,
        beta: Optional[float] = None,
    ):
        r"""Initialize similarity metric.

        Args:
            source: Source image from which to compute ``norm``. If ``None``, only use ``target`` if specified.
            target: Target image from which to compute ``norm``. If ``None``, only use ``source`` if specified.
            norm: Positive factor by which to divide loss. If ``None`` or ``True``, use ``source`` and/or ``target``.
                If ``False`` or both ``source`` and ``target`` are ``None``, a normalization factor of one is used.
            delta: Specifies the threshold at which to change between delta-scaled L1 and L2 loss.
            beta: Alternative name for ``delta`` to be compatible with ``SmoothL1ImageLoss``.

        """
        if beta is not None:
            if delta is not None:
                raise ValueError(
                    f"{type(self).__name__}() 'delta' and 'beta' are mutually exclusive"
                )
            delta = beta
        elif delta is None:
            delta = 1.0
        super().__init__(source, target, norm)
        self.delta = delta

    @property
    def beta(self) -> float:
        return self.delta

    def forward(self, source: Tensor, target: Tensor, mask: Optional[Tensor] = None) -> Tensor:
        r"""Evaluate image dissimilarity loss."""
        return L.huber_loss(source, target, mask=mask, norm=self.norm, delta=self.delta)


class SmoothL1ImageLoss(NormalizedPairwiseImageLoss):
    r"""Average smooth L1 loss."""

    def __init__(
        self,
        source: Optional[Tensor] = None,
        target: Optional[Tensor] = None,
        norm: Optional[Union[bool, Tensor]] = None,
        beta: Optional[float] = None,
        delta: Optional[float] = None,
    ):
        r"""Initialize similarity metric.

        Args:
            source: Source image from which to compute ``norm``. If ``None``, only use ``target`` if specified.
            target: Target image from which to compute ``norm``. If ``None``, only use ``source`` if specified.
            norm: Positive factor by which to divide loss. If ``None`` or ``True``, use ``source`` and/or ``target``.
                If ``False`` or both ``source`` and ``target`` are ``None``, a normalization factor of one is used.
            beta: Specifies the threshold at which to change between delta-scaled L1 and L2 loss.
            delta: Alternative name for ``beta`` to be compatible with ``HuberImageLoss``.

        """
        if delta is not None:
            if beta is not None:
                raise ValueError(
                    f"{type(self).__name__}() 'beta' and 'detla' are mutually exclusive"
                )
            beta = delta
        elif beta is None:
            beta = 1.0
        super().__init__(source, target, norm)
        self.beta = beta

    @property
    def delta(self) -> float:
        return self.beta

    def forward(self, source: Tensor, target: Tensor, mask: Optional[Tensor] = None) -> Tensor:
        r"""Evaluate image dissimilarity loss."""
        return L.smooth_l1_loss(source, target, mask=mask, norm=self.norm)


class L2ImageLoss(NormalizedPairwiseImageLoss):
    r"""Average squared intensity differences."""

    def forward(self, source: Tensor, target: Tensor, mask: Optional[Tensor] = None) -> Tensor:
        r"""Evaluate image dissimilarity loss."""
        return L.mse_loss(source, target, mask=mask, norm=self.norm)


MSE = L2ImageLoss


class SSD(NormalizedPairwiseImageLoss):
    r"""Sum of squared intensity differences."""

    def forward(self, source: Tensor, target: Tensor, mask: Optional[Tensor] = None) -> Tensor:
        r"""Evaluate image dissimilarity loss."""
        return L.ssd_loss(source, target, mask=mask, norm=self.norm)


class MI(PairwiseImageLoss):
    r"""Mutual information loss using Parzen window estimate with Gaussian kernel."""

    def __init__(
        self,
        vmin: Optional[float] = None,
        vmax: Optional[float] = None,
        bins: Optional[int] = None,
        sample: Optional[float] = None,
        num_bins: Optional[int] = None,
        num_samples: Optional[int] = None,
        sample_ratio: Optional[float] = None,
        normalized: bool = False,
    ):
        r"""Initialize mutual information loss term.

        See :func:`deepali.losses.functional.mi_loss`.

        """

        if bins is not None:
            if num_bins is not None:
                raise ValueError(
                    f"{type(self).__name__}() 'bins' and 'num_bins' are mutually exclusive"
                )
            num_bins = bins

        if sample is not None:
            if sample_ratio is not None or num_samples is not None:
                raise ValueError(
                    f"{type(self).__name__}() 'sample', 'sample_ratio', and 'num_samples' are mutually exclusive"
                )
            if 0 < sample < 1:
                sample_ratio = float(sample)
            else:
                try:
                    num_samples = int(sample)
                except TypeError:
                    pass
                if num_samples is None or float(num_samples) != sample:
                    raise ValueError(
                        f"{type(self).__name__}() 'sample' must be float in (0, 1) or positive int"
                    )
        if num_samples == -1:
            num_samples = None
        if num_samples is not None and (not isinstance(num_samples, int) or num_samples <= 0):
            raise ValueError(
                f"{type(self).__name__}() 'num_samples' must be positive integral value"
            )
        if sample_ratio is not None and (sample_ratio <= 0 or sample_ratio >= 1):
            raise ValueError(
                f"{type(self).__name__}() 'sample_ratio' must be in closed interval [0, 1]"
            )

        super().__init__()
        self.vmin = vmin
        self.vmax = vmax
        self.num_bins = num_bins
        self.num_samples = num_samples
        self.sample_ratio = sample_ratio
        self._normalized = normalized

    @property
    def bins(self) -> int:
        return self.num_bins

    @property
    def normalized(self) -> bool:
        return self._normalized

    def forward(self, source: Tensor, target: Tensor, mask: Optional[Tensor] = None) -> Tensor:
        r"""Evaluate patch dissimilarity loss."""
        return L.mi_loss(
            source,
            target,
            mask=mask,
            vmin=self.vmin,
            vmax=self.vmax,
            num_bins=self.num_bins,
            num_samples=self.num_samples,
            sample_ratio=self.sample_ratio,
            normalized=self.normalized,
        )

    def extra_repr(self) -> str:
        return (
            f"vmin={self.vmin!r}, "
            f"vmax={self.vmax!r}, "
            f"num_bins={self.num_bins!r}, "
            f"num_samples={self.num_samples!r}, "
            f"sampling_ratio={self.sample_ratio!r}, "
            f"normalized={self.normalized!r}"
        )


class NMI(MI):
    r"""Normalized mutual information loss using Parzen window estimate with Gaussian kernel."""

    def __init__(
        self,
        vmin: Optional[float] = None,
        vmax: Optional[float] = None,
        bins: Optional[int] = None,
        sample: Optional[float] = None,
        num_bins: Optional[int] = None,
        num_samples: Optional[int] = None,
        sample_ratio: Optional[float] = None,
    ):
        r"""Initialize normalized mutual information loss term.

        See :func:`deepali.losses.functional.nmi_loss`.

        """
        super().__init__(
            vmin=vmin,
            vmax=vmax,
            bins=bins,
            sample=sample,
            num_bins=num_bins,
            num_samples=num_samples,
            sample_ratio=sample_ratio,
        )

    def extra_repr(self) -> str:
        return (
            f"vmin={self.vmin!r}, "
            f"vmax={self.vmax!r}, "
            f"num_bins={self.num_bins!r}, "
            f"num_samples={self.num_samples!r}, "
            f"sampling_ratio={self.sample_ratio!r}"
        )


class PatchwiseImageLoss(PairwiseImageLoss):
    r"""Pairwise similarity of 2D image patches defined within a 3D volume."""

    def __init__(self, patches: Tensor, loss_fn: PairwiseImageLoss = SSD()):
        r"""Initialize loss term.

        Args:
            patches: Patch sampling points as tensor of shape ``(N, Z, Y, X, 3)``.
            loss_fn: Pairwise image similarity loss term used to evaluate similarity of patches.

        """
        super().__init__()
        if not isinstance(patches, Tensor):
            raise TypeError("PatchwiseImageLoss() 'patches' must be Tensor")
        if not patches.ndim == 5 or patches.shape[-1] != 3:
            raise ValueError("PatchwiseImageLoss() 'patches' must have shape (N, Z, Y, X, 3)")
        self.patches = patches
        self.loss_fn = loss_fn

    def forward(self, source: Tensor, target: Tensor, mask: Optional[Tensor] = None) -> Tensor:
        r"""Evaluate patch dissimilarity loss."""
        if target.ndim != 5:
            raise ValueError(
                "PatchwiseImageLoss.forward() 'target' must have shape (N, C, Z, Y, X)"
            )
        if source.shape != target.shape:
            raise ValueError(
                "PatchwiseImageLoss.forward() 'source' must have same shape as 'target'"
            )
        if mask is not None:
            if mask.shape != target.shape:
                raise ValueError(
                    "PatchwiseImageLoss.forward() 'mask' must have same shape as 'target'"
                )
            mask = self._reshape(U.grid_sample_mask(mask, self.patches))
        source = self._reshape(U.grid_sample(source, self.patches))
        target = self._reshape(U.grid_sample(target, self.patches))
        return self.loss_fn(source, target, mask=mask)

    @staticmethod
    def _reshape(x: Tensor) -> Tensor:
        r"""Reshape tensor to (N * Z, C, 1, Y, X) such that each patch is a separate image in the batch."""
        N, C, Z, Y, X = x.shape
        x = x.permute(0, 2, 1, 3, 4)  # N, Z, C, Y, X
        x = x.reshape(N * Z, C, 1, Y, X)
        return x


PatchLoss = PatchwiseImageLoss



r"""Auxiliary functions for reading and writing image files."""

from typing import Tuple

from torch import Tensor

from deepali.core.grid import Grid
from deepali.core.pathlib import PathUri

from .meta import has_meta_image_suffix, read_meta_image, write_meta_image
from .nifti import has_nifti_image_suffix, read_nifti_image, write_nifti_image


def read_image(path: PathUri) -> Tuple[Tensor, Grid]:
    r"""Read image data from specified file path.

    Args:
        path: Input file path.

    Returns:
        data: Image data tensor.
        grid: Image sampling grid.

    """
    if has_meta_image_suffix(path):
        return read_meta_image(path)
    if has_nifti_image_suffix(path):
        return read_nifti_image(path)
    try:
        from .sitk import read_sitk_image
    except ImportError:
        raise RuntimeError(
            f"Cannot read image {path}."
            " Image file formats other than MetaImage and NIfTI require SimpleITK to be installed."
        )
    return read_sitk_image(path)


def write_image(data: Tensor, grid: Grid, path: PathUri, compress: bool = True) -> None:
    r"""Write image data to specified file path.

    Args:
        data: Image data tensor.
        grid: Image sampling grid.

    """
    if has_meta_image_suffix(path):
        return write_meta_image(data, grid, path, compress=compress)
    if has_nifti_image_suffix(path):
        return write_nifti_image(data, grid, path)
    try:
        from .sitk import write_sitk_image
    except ImportError:
        raise RuntimeError(
            f"Cannot write image {path}."
            " Image file formats other than MetaImage and NIfTI require SimpleITK to be installed."
        )
    return write_sitk_image(data, grid, path, compress=compress)



from typing import Optional, Sequence

import pytest
import torch
import torch.nn.functional as F
from torch import Tensor
from torch.random import Generator

from deepali.core import Grid
from deepali.core.enum import FlowDerivativeKeys
import deepali.core.bspline as B
import deepali.core.functional as U


PERIODIC_FLOW_X_SCALE = 2 * torch.pi
PERIODIC_FLOW_U_SCALE = 0.1


def periodic_flow(p: Tensor) -> Tensor:
    q = p.mul(PERIODIC_FLOW_X_SCALE)
    u = q.narrow(1, 0, 1).sin().neg_()
    v = q.narrow(1, 1, 1).cos()
    f = torch.cat([u, v], dim=1)
    return f.mul_(PERIODIC_FLOW_U_SCALE)


def periodic_flow_du_dx(p: Tensor) -> Tensor:
    q = p.mul(PERIODIC_FLOW_X_SCALE)
    g = q.narrow(1, 0, 1).cos()
    g = g.mul_(-PERIODIC_FLOW_X_SCALE * PERIODIC_FLOW_U_SCALE)
    return g


def periodic_flow_du_dy(p: Tensor) -> Tensor:
    return torch.zeros((p.shape[0], 1) + p.shape[2:], dtype=p.dtype, device=p.device)


def periodic_flow_du_dxx(p: Tensor) -> Tensor:
    q = p.mul(PERIODIC_FLOW_X_SCALE)
    g = q.narrow(1, 0, 1).sin()
    g = g.mul_(PERIODIC_FLOW_X_SCALE * PERIODIC_FLOW_X_SCALE * PERIODIC_FLOW_U_SCALE)
    return g


def periodic_flow_du_dyy(p: Tensor) -> Tensor:
    return torch.zeros((p.shape[0], 1) + p.shape[2:], dtype=p.dtype, device=p.device)


def periodic_flow_dv_dx(p: Tensor) -> Tensor:
    return torch.zeros((p.shape[0], 1) + p.shape[2:], dtype=p.dtype, device=p.device)


def periodic_flow_dv_dy(p: Tensor) -> Tensor:
    q = p.mul(PERIODIC_FLOW_X_SCALE)
    g = q.narrow(1, 1, 1).sin()
    g = g.mul_(-PERIODIC_FLOW_X_SCALE * PERIODIC_FLOW_U_SCALE)
    return g


def periodic_flow_dv_dxx(p: Tensor) -> Tensor:
    return torch.zeros((p.shape[0], 1) + p.shape[2:], dtype=p.dtype, device=p.device)


def periodic_flow_dv_dyy(p: Tensor) -> Tensor:
    q = p.mul(PERIODIC_FLOW_X_SCALE)
    g = q.narrow(1, 1, 1).cos()
    g = g.mul_(-PERIODIC_FLOW_X_SCALE * PERIODIC_FLOW_X_SCALE * PERIODIC_FLOW_U_SCALE)
    return g


def periodic_flow_deriv(p: Tensor, which: str) -> Tensor:
    deriv_fn = {
        "du/dx": periodic_flow_du_dx,
        "du/dy": periodic_flow_du_dy,
        "dv/dx": periodic_flow_dv_dx,
        "dv/dy": periodic_flow_dv_dy,
        "du/dxx": periodic_flow_du_dxx,
        "du/dyy": periodic_flow_du_dyy,
        "dv/dxx": periodic_flow_dv_dxx,
        "dv/dyy": periodic_flow_dv_dyy,
    }
    return deriv_fn[which](p)


def periodic_flow_divergence(p: Tensor) -> Tensor:
    du_dx = periodic_flow_du_dx(p)
    dv_dy = periodic_flow_dv_dy(p)
    return du_dx.add(dv_dy)


def random_svf(
    size: Sequence[int],
    stride: int = 1,
    generator: Optional[Generator] = None,
) -> Tensor:
    cp_grid_size = B.cubic_bspline_control_point_grid_size(size, stride=stride)
    data = torch.randn((1, 3) + cp_grid_size, generator=generator)
    data = U.fill_border(data, margin=3, value=0, inplace=True)
    return B.evaluate_cubic_bspline(data, size=size, stride=stride)


def difference(a: Tensor, b: Tensor, margin: int = 0) -> Tensor:
    assert a.shape == b.shape
    i = [
        slice(0, n, 1) if dim < 2 else slice(margin, n - margin, 1) for dim, n in enumerate(a.shape)
    ]
    return a[i].sub(b[i])


def test_flow_curl() -> None:
    # 2-dimensional vector field
    p = U.move_dim(Grid(size=(32, 24)).coords().unsqueeze_(0), -1, 1)

    x = p.narrow(1, 0, 1)
    y = p.narrow(1, 1, 1)

    flow = torch.cat([y.neg(), x], dim=1)

    curl = U.curl(flow)
    assert isinstance(curl, Tensor)
    assert curl.shape == (flow.shape[0], 1) + flow.shape[2:]
    assert curl.dtype == flow.dtype
    assert curl.device == flow.device
    assert curl.sub(2).abs().lt(1e-6).all()

    # 3-dimensional vector field
    p = U.move_dim(Grid(size=(64, 32, 16)).coords().unsqueeze_(0), -1, 1)

    x = p.narrow(1, 0, 1)
    y = p.narrow(1, 1, 1)
    z = p.narrow(1, 2, 1)

    flow = torch.cat([x.mul(z), y.mul(z), x.mul(y)], dim=1)

    curl = U.curl(flow)
    assert isinstance(curl, Tensor)
    assert curl.shape == flow.shape
    assert curl.dtype == flow.dtype
    assert curl.device == flow.device

    expected = x.sub(y)
    expected = torch.cat([expected, expected, torch.zeros_like(expected)], dim=1)
    error = difference(curl, expected, margin=1)
    assert error.abs().max().lt(1e-5)

    div = U.divergence(curl)
    assert div.abs().max().lt(1e-5)


def test_flow_derivatives() -> None:
    # 2D vector field defined by periodic functions
    p = U.move_dim(Grid(size=(120, 100)).coords().unsqueeze_(0), -1, 1)

    flow = periodic_flow(p)

    which = FlowDerivativeKeys.all(spatial_dims=2, order=1)
    which.append("du/dxx")
    which.append("dv/dyy")

    deriv = U.flow_derivatives(flow, which=which)
    assert isinstance(deriv, dict)
    assert all(isinstance(k, str) for k in deriv.keys())
    assert all(isinstance(v, Tensor) for v in deriv.values())
    assert all(v.shape == (flow.shape[0], 1) + flow.shape[2:] for v in deriv.values())

    for key, value in deriv.items():
        expected = periodic_flow_deriv(p, key)
        order = FlowDerivativeKeys.order(key)
        dif = difference(value, expected, margin=order)
        tol = 0.003 * (10 ** (order - 1))
        assert dif.abs().max().lt(tol), f"flow derivative {key}"

    # 3D vector field
    p = U.move_dim(Grid(size=(64, 32, 16)).coords().unsqueeze_(0), -1, 1)

    x = p.narrow(1, 0, 1)
    y = p.narrow(1, 1, 1)
    z = p.narrow(1, 2, 1)

    flow = torch.cat([x.mul(z), y.mul(z), x.mul(y)], dim=1)
    deriv = U.flow_derivatives(flow, order=1)

    assert difference(deriv["du/dx"], z).abs().max().lt(1e-5)
    assert deriv["du/dy"].abs().max().lt(1e-5)
    assert difference(deriv["du/dz"], x).abs().max().lt(1e-5)

    assert deriv["dv/dx"].abs().max().lt(1e-5)
    assert difference(deriv["dv/dy"], z).abs().max().lt(1e-5)
    assert difference(deriv["dv/dz"], y).abs().max().lt(1e-5)

    assert difference(deriv["dw/dx"], y).abs().max().lt(1e-5)
    assert difference(deriv["dw/dy"], x).abs().max().lt(1e-5)
    assert deriv["dw/dz"].abs().max().lt(1e-5)


def test_flow_divergence() -> None:
    grid = Grid(size=(16, 14))
    offset = U.translation([0.1, 0.2]).unsqueeze_(0)
    flow = U.affine_flow(offset, grid)
    div = U.divergence(flow)
    assert isinstance(div, Tensor)
    assert div.shape == (flow.shape[0], 1) + flow.shape[2:]
    assert div.abs().max().lt(1e-5)

    points = U.move_dim(Grid(size=(64, 64)).coords().unsqueeze_(0), -1, 1)
    flow = periodic_flow(points)
    div = U.divergence(flow)
    assert isinstance(div, Tensor)
    assert div.shape == (flow.shape[0], 1) + flow.shape[2:]
    expected = periodic_flow_divergence(points)
    dif = difference(div, expected, margin=1)
    assert dif.abs().max().lt(0.01)

    p = U.move_dim(Grid(size=(64, 32, 16)).coords().unsqueeze_(0), -1, 1)
    x = p.narrow(1, 0, 1)
    y = p.narrow(1, 1, 1)
    z = p.narrow(1, 2, 1)
    flow = torch.cat([x.mul(z), y.mul(z), x.mul(y)], dim=1)
    div = U.divergence(flow)
    assert isinstance(div, Tensor)
    assert div.shape == (flow.shape[0], 1) + flow.shape[2:]
    error = difference(div, z.mul(2))
    assert error.abs().max().lt(1e-5)


def test_flow_divergence_free() -> None:
    data = torch.randn((1, 1, 16, 24)).mul_(0.01)
    flow = U.divergence_free_flow(data)
    assert flow.shape == (data.shape[0], 2) + data.shape[2:]
    div = U.divergence(flow)
    assert div.abs().max().lt(1e-5)

    data = torch.randn((3, 2, 16, 24, 32)).mul_(0.01)
    flow = U.divergence_free_flow(data, sigma=2.0)
    assert flow.shape == (data.shape[0], 3) + data.shape[2:]
    div = U.divergence(flow)
    assert div[0, 0, 1:-1, 1:-1, 1:-1].abs().max().lt(1e-4)

    coef = F.pad(data, (1, 2, 1, 2, 1, 2))
    flow = U.divergence_free_flow(coef, mode="bspline", sigma=0.8)
    assert flow.shape == (data.shape[0], 3) + data.shape[2:]
    div = U.divergence(flow)
    assert div[0, 0, 1:-1, 1:-1, 1:-1].abs().max().lt(1e-4)

    # constructing a divergence-free field using curl() seems to work best given
    # the higher magnitude and no need for Gaussian blurring of the random field
    # where each component is sampled i.i.d. from a normal distribution
    data = torch.randn((5, 3, 16, 24, 32)).mul_(0.2)
    flow = U.divergence_free_flow(data)
    assert flow.shape == data.shape
    div = U.divergence(flow)
    assert div.abs().max().lt(1e-4)


def test_flow_jacobian() -> None:
    # 2D flow field
    p = Grid(size=(64, 32)).coords(channels_last=False).unsqueeze_(0)

    x = p.narrow(1, 0, 1)
    y = p.narrow(1, 1, 1)

    interior = [slice(1, n - 1) for n in p.shape[2:]]

    # u = [x^2, xy]
    flow = torch.cat([x.square(), x.mul(y)], dim=1)

    jac = torch.zeros((p.shape[0],) + p.shape[2:] + (2, 2))
    jac[..., 0, 0] = x.squeeze(1).mul(2)
    jac[..., 1, 0] = y.squeeze(1)
    jac[..., 1, 1] = x.squeeze(1)

    derivs = U.jacobian_dict(flow)
    for (i, j), deriv in derivs.items():
        atol = 1e-5
        error = difference(jac[..., i, j].unsqueeze(1), deriv)
        if (i, j) == (0, 0):
            error = error[[slice(None), slice(None)] + interior]
        if error.abs().max().gt(atol):
            raise AssertionError(f"max absolute difference of jac[{i}, {j}] > {atol}")

    mat = U.jacobian_matrix(flow)
    assert torch.allclose(
        mat[[slice(None)] + interior],
        jac[[slice(None)] + interior],
        atol=1e-5,
    )

    jac[..., 0, 0] += 1
    jac[..., 1, 1] += 1

    mat = U.jacobian_matrix(flow, add_identity=True)
    assert torch.allclose(
        mat[[slice(None)] + interior],
        jac[[slice(None)] + interior],
        atol=1e-5,
    )

    det = U.jacobian_det(flow)
    assert torch.allclose(
        det[[slice(None), 0] + interior],
        jac[[slice(None)] + interior].det(),
        atol=1e-5,
    )

    # 3D flow field
    p = Grid(size=(64, 32, 16)).coords(channels_last=False).unsqueeze_(0)

    x = p.narrow(1, 0, 1)
    y = p.narrow(1, 1, 1)
    z = p.narrow(1, 2, 1)

    interior = [slice(1, n - 1) for n in p.shape[2:]]

    # u = [z^2, 0, xy]
    flow = torch.cat([z.square(), torch.zeros_like(y), x.mul(y)], dim=1)

    jac = torch.zeros((p.shape[0],) + p.shape[2:] + (3, 3))
    jac[..., 0, 2] = z.squeeze(1).mul(2)
    jac[..., 2, 0] = y.squeeze(1)
    jac[..., 2, 1] = x.squeeze(1)

    derivs = U.jacobian_dict(flow)
    for (i, j), deriv in derivs.items():
        atol = 1e-5
        error = difference(jac[..., i, j].unsqueeze(1), deriv)
        if (i, j) == (0, 2):
            error = error[[slice(None), slice(None)] + interior]
        if error.abs().max().gt(atol):
            raise AssertionError(f"max absolute difference of jac[{i}, {j}] > {atol}")

    mat = U.jacobian_matrix(flow)
    assert torch.allclose(
        mat[[slice(None)] + interior],
        jac[[slice(None)] + interior],
        atol=1e-5,
    )

    jac[..., 0, 0] += 1
    jac[..., 1, 1] += 1
    jac[..., 2, 2] += 1

    mat = U.jacobian_matrix(flow, add_identity=True)
    assert torch.allclose(
        mat[[slice(None)] + interior],
        jac[[slice(None)] + interior],
        atol=1e-5,
    )

    det = U.jacobian_det(flow)
    assert torch.allclose(
        det[[slice(None), 0] + interior],
        jac[[slice(None)] + interior].det(),
        atol=1e-5,
    )

    # u = [0, x + y^3, yz]
    flow = torch.cat([torch.zeros_like(x), x.add(y.pow(3)), y.mul(z)], dim=1)

    jac = torch.zeros((p.shape[0],) + p.shape[2:] + (3, 3))
    jac[..., 1, 0] = 1
    jac[..., 1, 1] = y.squeeze(1).square().mul(3)
    jac[..., 2, 1] = z.squeeze(1)
    jac[..., 2, 2] = y.squeeze(1)

    derivs = U.jacobian_dict(flow)
    for (i, j), deriv in derivs.items():
        atol = 1e-5
        error = difference(jac[..., i, j].unsqueeze(1), deriv)
        if (i, j) == (1, 1):
            atol = 0.005
            error = error[[slice(None), slice(None)] + interior]
        if error.abs().max().gt(atol):
            raise AssertionError(f"max absolute difference of jac[{i}, {j}] > {atol}")

    mat = U.jacobian_matrix(flow)
    assert torch.allclose(
        mat[[slice(None)] + interior],
        jac[[slice(None)] + interior],
        atol=0.005,
    )

    jac[..., 0, 0] += 1
    jac[..., 1, 1] += 1
    jac[..., 2, 2] += 1

    mat = U.jacobian_matrix(flow, add_identity=True)
    assert torch.allclose(
        mat[[slice(None)] + interior],
        jac[[slice(None)] + interior],
        atol=0.005,
    )

    det = U.jacobian_det(flow)
    assert torch.allclose(
        det[[slice(None), 0] + interior],
        jac[[slice(None)] + interior].det(),
        atol=0.01,
    )


def test_flow_lie_bracket() -> None:
    p = U.move_dim(Grid(size=(64, 32, 16)).coords().unsqueeze_(0), -1, 1)

    x = p.narrow(1, 0, 1)
    y = p.narrow(1, 1, 1)
    z = p.narrow(1, 2, 1)

    # u = [yz, xz, xy] and v = [x, y, z]
    u = torch.cat([y.mul(z), x.mul(z), x.mul(y)], dim=1)
    v = torch.cat([x, y, z], dim=1)
    w = u

    lb_uv = U.lie_bracket(u, v)
    assert torch.allclose(U.lie_bracket(v, u), lb_uv.neg())
    assert U.lie_bracket(u, u).abs().lt(1e-6).all()
    assert torch.allclose(lb_uv, w, atol=1e-6)

    # u = [z^2, 0, xy] and v = [0, x + y^3, yz]
    u = torch.cat([z.square(), torch.zeros_like(y), x.mul(y)], dim=1)
    v = torch.cat([torch.zeros_like(x), x.add(y.pow(3)), y.mul(z)], dim=1)
    w = torch.cat([-2 * y * z**2, z**2, x * y**2 - x**2 - x * y**3], dim=1).neg_()

    lb_uv = U.lie_bracket(u, v)
    assert torch.allclose(U.lie_bracket(v, u), lb_uv.neg())
    assert U.lie_bracket(u, u).abs().lt(1e-6).all()
    error = difference(lb_uv, w).abs()
    assert error[:, :, 1:-1, 1:-1, 1:-1].max().lt(1e-5)
    assert error.max().lt(0.134)


def test_flow_logv() -> None:
    size = (128, 128, 128)
    generator = torch.Generator().manual_seed(42)
    v = random_svf(size, stride=8, generator=generator).mul_(0.1)
    u = U.expv(v)
    w = U.logv(u)
    error = w.sub(v).norm(dim=1, keepdim=True)
    assert error.mean().lt(0.001)
    assert error.max().lt(0.02)


def test_flow_compose_svfs() -> None:
    # 3D flow fields
    p = U.move_dim(Grid(size=(64, 32, 16)).coords().unsqueeze_(0), -1, 1)

    x = p.narrow(1, 0, 1)
    y = p.narrow(1, 1, 1)
    z = p.narrow(1, 2, 1)

    with pytest.raises(ValueError):
        U.compose_svfs(p, p, bch_terms=-1)
    with pytest.raises(NotImplementedError):
        U.compose_svfs(p, p, bch_terms=6)

    # u = [yz, xz, xy] and v = u
    u = v = torch.cat([y.mul(z), x.mul(z), x.mul(y)], dim=1)

    w = U.compose_svfs(u, v, bch_terms=0)
    assert torch.allclose(w, u.add(v))
    w = U.compose_svfs(u, v, bch_terms=1)
    assert torch.allclose(w, u.add(v))
    w = U.compose_svfs(u, v, bch_terms=2)
    assert torch.allclose(w, u.add(v))
    w = U.compose_svfs(u, v, bch_terms=3)
    assert torch.allclose(w, u.add(v))
    w = U.compose_svfs(u, v, bch_terms=4)
    assert torch.allclose(w, u.add(v), atol=1e-5)
    w = U.compose_svfs(u, v, bch_terms=5)
    assert torch.allclose(w, u.add(v), atol=1e-5)

    # u = [yz, xz, xy] and v = [x, y, z]
    u = torch.cat([y.mul(z), x.mul(z), x.mul(y)], dim=1)
    v = torch.cat([x, y, z], dim=1)

    w = U.compose_svfs(u, v, bch_terms=0)
    assert torch.allclose(w, u.add(v))
    w = U.compose_svfs(u, v, bch_terms=1)
    assert torch.allclose(w, u.mul(0.5).add(v), atol=1e-6)

    # u = random_svf(), u -> 0 at boundary
    # v = random_svf(), v -> 0 at boundary
    size = (64, 64, 64)
    generator = torch.Generator().manual_seed(42)
    u = random_svf(size, stride=4, generator=generator).mul_(0.1)
    v = random_svf(size, stride=4, generator=generator).mul_(0.05)
    w = U.compose_svfs(u, v, bch_terms=5)

    flow_u = U.expv(u)
    flow_v = U.expv(v)
    flow_w = U.expv(w)
    flow = U.compose_flows(flow_u, flow_v)

    error = flow_w.sub(flow).norm(dim=1)
    assert error.max().lt(0.01)



r"""Functions for dealing with dataset sample collections."""

from copy import copy as shallowcopy
from collections import OrderedDict
from dataclasses import fields, is_dataclass
from typing import Any, Mapping, Sequence, Tuple

from deepali.core.typing import Sample, is_namedtuple


__all__ = (
    "Sample",
    "sample_field_names",
    "sample_field_value",
    "replace_all_sample_field_values",
)


def sample_field_names(sample: Sample) -> Tuple[str]:
    r"""Get names of fields in data sample."""
    if is_dataclass(sample):
        return tuple((field.name for field in fields(sample)))
    if is_namedtuple(sample):
        return sample._fields
    if not isinstance(sample, Mapping):
        raise TypeError("Dataset 'sample' must be dataclass, Mapping, or NamedTuple")
    return tuple(sample.keys())


def sample_field_value(sample: Sample, name: str) -> Any:
    r"""Get sample value of named data field."""
    if isinstance(sample, Mapping):
        return sample[name]
    return getattr(sample, name)


def replace_all_sample_field_values(sample: Sample, values: Sequence[Any]) -> Sample:
    r"""Replace all sample field values."""
    names = sample_field_names(sample)
    if len(names) != len(values):
        raise ValueError("replace_all_values() 'values' must contain an entry for every field")
    if is_dataclass(sample):
        result = shallowcopy(sample)
        for name, value in zip(names, values):
            setattr(result, name, value)
        return result
    if is_namedtuple(sample):
        return sample._replace(**{name: value for name, value in zip(names, values)})
    if isinstance(sample, OrderedDict):
        return OrderedDict([(name, value) for name, value in zip(names, values)])
    return {name: value for name, value in zip(names, values)}



r"""Building blocks to construct subnetworks, and predefined subnetworks (blocks)."""

from .residual import ResidualUnit

from .skip import DenseBlock
from .skip import SkipConnection, Shortcut


# fmt: off
__all__ = (
    "DenseBlock",
    "ResidualUnit",
    "Shortcut",
    "SkipConnection",
)
# fmt: on



r"""Fully connected layers."""

import math
from typing import Optional

from torch.nn import init
from torch.nn.modules import Linear as _Linear


class Linear(_Linear):
    r"""Fully connected layer."""

    def __init__(
        self,
        in_features: int,
        out_features: int,
        bias: bool = True,
        init: Optional[str] = "uniform",
    ) -> None:
        self.bias_init = "uniform" if isinstance(bias, bool) else bias
        self.weight_init = init
        super().__init__(in_features, out_features, bias=bool(bias))

    def reset_parameters(self) -> None:
        # Initialize weights
        if self.weight_init == "uniform":
            init.kaiming_uniform_(self.weight, a=math.sqrt(5))
        elif self.weight_init == "constant":
            init.constant_(self.weight, 0.1)
        elif self.weight_init == "zeros":
            init.constant_(self.weight, 0.0)
        else:
            raise AssertionError(
                "Linear.reset_parameters() invalid 'init' value: {self.weight_init!r}"
            )
        # Initialize bias
        if self.bias is not None:
            if self.bias_init == "uniform":
                fan_in, _ = init._calculate_fan_in_and_fan_out(self.weight)
                bound = 1 / math.sqrt(fan_in)
                init.uniform_(self.bias, -bound, bound)
            elif self.bias_init == "constant":
                init.constant_(self.bias, 0.1)
            elif self.bias_init == "zeros":
                init.constant_(self.bias, 0.0)
            else:
                raise AssertionError(
                    "Linear.reset_parameters() invalid 'bias' value: {self.bias_init!r}"
                )



import pytest

import torch
from torch import Tensor

from deepali.core import bspline as B


def test_cubic_bspline_interpolation_weights() -> None:
    kernel = B.cubic_bspline_interpolation_weights(5)
    assert isinstance(kernel, Tensor)
    assert kernel.shape == (5, 4)

    assert torch.allclose(kernel, B.cubic_bspline_interpolation_weights(5, derivative=0))

    kernels = B.cubic_bspline_interpolation_weights(5, derivative=[0, 1])
    assert isinstance(kernels, tuple)
    assert len(kernels) == 2
    assert torch.allclose(kernels[0], B.cubic_bspline_interpolation_weights(5, derivative=0))
    assert torch.allclose(kernels[1], B.cubic_bspline_interpolation_weights(5, derivative=1))

    with pytest.raises(ValueError):
        B.cubic_bspline_interpolation_weights([5], derivative=[0, 1])
    with pytest.raises(ValueError):
        B.cubic_bspline_interpolation_weights([5, 5], derivative=[0])


def test_cubic_bspline_interpolation_at_control_points() -> None:
    r"""Test evaluation of cubic B-spline at control point grid locations only."""

    kernel_all = B.cubic_bspline_interpolation_weights(5)
    kernel_cps = B.cubic_bspline_interpolation_weights(1)
    assert kernel_all.shape == (5, 4)
    assert kernel_cps.shape == (1, 4)
    assert torch.allclose(kernel_cps, kernel_all[0])

    # Single non-zero control point coefficient
    data = torch.zeros((1, 1, 4))
    data[0, 0, 1] = 1

    values_all = B.evaluate_cubic_bspline(data, stride=5, kernel=kernel_all)
    values_cps = B.evaluate_cubic_bspline(data, stride=1, kernel=kernel_cps)
    assert torch.allclose(B.evaluate_cubic_bspline(data, stride=1), values_cps)

    assert values_all.shape == (1, 1, 5)
    assert values_cps.shape == (1, 1, 1)

    assert torch.allclose(values_cps, values_all[:, :, ::5])
    assert torch.allclose(values_cps[0, 0], kernel_cps[0, 1])

    # Random uniformly distributed control point coefficients
    data = torch.rand((1, 1, 35))

    values_all = B.evaluate_cubic_bspline(data, stride=5, kernel=kernel_all)
    values_cps = B.evaluate_cubic_bspline(data, stride=1, kernel=kernel_cps)
    assert torch.allclose(B.evaluate_cubic_bspline(data, stride=1), values_cps)

    assert values_all.shape == (1, 1, 32 * 5)
    assert values_cps.shape == (1, 1, 32)
    assert torch.allclose(values_cps, values_all[:, :, ::5])



r"""Linear homogeneous coordinate transformations."""

import re
from typing import Optional, Union

import torch
from torch import Tensor

from .linalg import homogeneous_transform
from .tensor import as_float_tensor, atleast_1d, cat_scalars
from .typing import Array, Device, DType, Scalar, Shape


__all__ = (
    "affine_rotation_matrix",
    "apply_transform",
    "euler_rotation_matrix",
    "euler_rotation_angles",
    "euler_rotation_order",
    "identity_transform",
    "rotation_matrix",
    "scaling_transform",
    "shear_matrix",
    "translation",
    "transform_points",
    "transform_vectors",
)


def apply_transform(transform: Tensor, points: Tensor, vectors: bool = False) -> Tensor:
    r"""Alias for :func:`.homogeneous_transform`."""
    return homogeneous_transform(transform, points, vectors=vectors)


def affine_rotation_matrix(matrix: Tensor) -> Tensor:
    r"""Get orthonormal rotation matrix from (homogeneous) affine transformation.

    This function assumes the following order of elementary transformations:
    1) Scaling, 2) Shearing, 3) Rotation, and 4) Translation.

    See also FullAffineTransform, AffineTransform, RigidTransform, etc.

    Args:
        matrix: Affine transformation as tensor of shape (..., 3, 3) or (..., 3, 4).

    Returns:
        Orthonormal rotation matrices with determinant 1 as tensor of shape (..., 3, 3).

    """
    # Translated from the C++ MIRTK code which is based on the book Graphics Gems:
    # https://github.com/BioMedIA/MIRTK/blob/6461e43d0ad0e0dcea2a65aeea213a78b420eae5/Modules/Numerics/src/Matrix.cc#L1446-L1548
    if not isinstance(matrix, Tensor):
        raise TypeError("affine_rotation_matrix() 'matrix' must be Tensor")
    if matrix.ndim < 2 or matrix.shape[-2] != 3 or matrix.shape[-1] not in (3, 4):
        raise ValueError("affine_rotation_matrix() 'matrix' must have shape (..., 3, 3|4)")
    matrix = matrix[..., :3].clone()
    # Compute X scale factor and normalize 1st column.
    sx: Tensor = torch.linalg.norm(matrix[..., 0], ord=2, dim=-1)
    matrix[..., 0] = matrix[..., 0].div(sx.unsqueeze(-1))
    # Compute XY shear factor and make 2nd column orthogonal to 1st.
    tansxy = matrix[..., 0].mul(matrix[..., 1]).sum(dim=-1)
    matrix[..., 1] = matrix[..., 1].sub(matrix[..., 0].mul(tansxy.unsqueeze(-1)))
    # Actually, tansxy and 2nd column are still to large by a factor of sy.
    # Now, compute Y scale and normalize 2nd column and rescale tansxy.
    sy: Tensor = torch.linalg.norm(matrix[..., 1], ord=2, dim=-1)
    matrix[..., 1] = matrix[..., 1].div(sy.unsqueeze(-1))
    tansxy = tansxy.div(sy)
    # Compute XZ and YZ shears, orthogonalize 3rd column.
    tansxz = matrix[..., 0].mul(matrix[..., 2]).sum(dim=-1)
    matrix[..., 2] = matrix[..., 2].sub(matrix[..., 0].mul(tansxz.unsqueeze(-1)))
    tansyz = matrix[..., 1].mul(matrix[..., 2]).sum(dim=-1)
    matrix[..., 2] = matrix[..., 2].sub(matrix[..., 1].mul(tansyz.unsqueeze(-1)))
    # Actually, tansxz, tansyz and 2nd column are still too large by a factor of sz.
    # Next, get Z scale, normalize 3rd column and scale tansxz and tansyz.
    sz: Tensor = torch.linalg.norm(matrix[..., 2], ord=2, dim=-1)
    matrix[..., 2] = matrix[..., 2].div(sz.unsqueeze(-1))
    tansxz = tansxz.div(sz)
    tansyz = tansyz.div(sz)
    # At this point, the columns are orthonormal. Check for a coordinate system flip.
    # If the determinant is -1, then negate the matrix (and the scaling factors).
    mask = matrix[..., 0].mul(matrix[..., 1].cross(matrix[..., 2], dim=-1)).sum(dim=-1).ge(0)
    mask = mask.unsqueeze(-1).unsqueeze(-1).expand_as(matrix)
    matrix = matrix.where(mask, -matrix)
    return matrix


def identity_transform(
    shape: Union[int, Shape],
    *args,
    homogeneous: bool = False,
    dtype: Optional[DType] = None,
    device: Optional[Device] = None,
) -> Tensor:
    r"""Create homogeneous coordinate transformation matrix of identity mapping.

    Args:
        shape: Shape of non-homogeneous point coordinates tensor ``(..., D)``, where the size of the
            last dimension corresponds to the number of spatial dimensions.
        homogeneous: Whether to return homogeneous transformation matrices.
        dtype: Data type of output matrix. If ``None``, use default dtype.
        device: Device on which to create matrix. If ``None``, use default device.

    Returns:
        If ``homogeneous=Falae``, a tensor of affine matrices of shape ``(..., D, D)`` is returned, and
        a tensor of homogeneous coordinate transformation matrices of shape ``(..., D, D + 1)``, otherwise.

    """
    shape_ = [int(n) for n in cat_scalars(shape, *args, device=device)]
    D = shape_[-1]
    J = tuple(range(D))
    matrix = torch.zeros(*shape_, D + 1 if homogeneous else D, dtype=dtype, device=device)
    matrix[..., J, J] = 1
    return matrix


def rotation_matrix(*args, **kwargs) -> Tensor:
    r"""Alias for :func:`.euler_rotation_matrix`."""
    return euler_rotation_matrix(*args, **kwargs)


def euler_rotation_matrix(
    angles: Union[Scalar, Array],
    order: Optional[str] = None,
    homogeneous: bool = False,
    dtype: Optional[DType] = None,
    device: Optional[Device] = None,
) -> Tensor:
    r"""Euler rotation matrices.

    Args:
        angles: Scalar rotation angle for a 2D rotation, or the three angles (alpha, beta, gamma)
            for an extrinsic rotation in the specified `order`. The argument can be a tensor of
            shape ``(..., D)``, with angles in the last dimension. All angles must be given in
            radians, and the first angle corresponds to the left-most rotation which is applied last.
        order: Order in which to compose elemental rotations. For example in 3D, "zxz" (or "ZXZ")
            means that the first rotation occurs about z by angle gamma, the second about x by
            angle beta, and the third rotation about z again by angle alpha. In 2D, this argument
            is ignored and a single rotation about z (plane normal) is applied. Alternatively,
            strings of the form "Rz o Rx o Rz" can be given as argument. When the first and last
            extrinsic rotation is about the same axis, the rotation is called proper, and the
            angles are referred to as proper Euler angles. When each elemental rotation in X, Y,
            and Z occurs exactly once, the angles are referred to as Tait-Bryan angles.
        homogeneous: Whether to return homogeneous transformation matrices.
        dtype: Data type of rotation matrix. If ``None``, use ``angles.dtype`` if it is
            a floating point type, and ``torch.float`` otherwise.
        device: Device on which to create rotation matrix. If ``None``, use ``angles.device``.

    Returns:
        Tensor of square rotation matrices of shape ``(..., D, D)`` if ``homogeneous=False``, or tensor
        of homogeneous coordinate transformation matrices of shape ``(..., D, D + 1)`` otherwise.
        If ``angles`` is a scalar or vector, a single rotation matrix is returned.

    See also:
        - https://mathworld.wolfram.com/EulerAngles.html
        - https://en.wikipedia.org/wiki/Euler_angles#Definition_by_extrinsic_rotations

    """
    if not isinstance(order, (str, type(None))):
        raise TypeError("euler_rotation_matrix() 'order' must be None or str")
    angles_ = atleast_1d(angles, dtype=dtype, device=device)
    angles_ = as_float_tensor(angles_)
    c = torch.cos(angles_)
    s = torch.sin(angles_)
    N = angles_.shape[-1]
    D = 2 if N == 1 else N
    order = euler_rotation_order(order, ndim=D)
    matrix = angles_.new_empty(angles_.shape[:-1] + (D, D + 1 if homogeneous else D))
    if homogeneous:
        matrix[..., D] = 0
    if D == 2:
        matrix[..., 0, 0] = c[..., 0]
        matrix[..., 0, 1] = -s[..., 0]
        matrix[..., 1, 0] = s[..., 0]
        matrix[..., 1, 1] = c[..., 0]
    elif D == 3:
        # See also the following Wikipedia page. Attention: Order of angles is reversed!
        # https://en.wikipedia.org/wiki/Euler_angles#Rotation_matrix
        if order == "XYZ":
            matrix[..., 0, 0] = c[..., 1] * c[..., 2]
            matrix[..., 0, 1] = -c[..., 1] * s[..., 2]
            matrix[..., 0, 2] = s[..., 1]
            matrix[..., 1, 0] = c[..., 0] * s[..., 2] + c[..., 2] * s[..., 0] * s[..., 1]
            matrix[..., 1, 1] = c[..., 0] * c[..., 2] - s[..., 0] * s[..., 1] * s[..., 2]
            matrix[..., 1, 2] = -c[..., 1] * s[..., 0]
            matrix[..., 2, 0] = s[..., 0] * s[..., 2] - c[..., 0] * c[..., 2] * s[..., 1]
            matrix[..., 2, 1] = c[..., 2] * s[..., 0] + c[..., 0] * s[..., 1] * s[..., 2]
            matrix[..., 2, 2] = c[..., 0] * c[..., 1]
        elif order == "ZYX":
            matrix[..., 0, 0] = c[..., 0] * c[..., 1]
            matrix[..., 0, 1] = c[..., 0] * s[..., 1] * s[..., 2] - c[..., 2] * s[..., 0]
            matrix[..., 0, 2] = s[..., 0] * s[..., 2] + c[..., 0] * c[..., 2] * s[..., 1]
            matrix[..., 1, 0] = c[..., 1] * s[..., 0]
            matrix[..., 1, 1] = c[..., 0] * c[..., 2] + s[..., 0] * s[..., 1] * s[..., 2]
            matrix[..., 1, 2] = c[..., 2] * s[..., 0] * s[..., 1] - c[..., 0] * s[..., 2]
            matrix[..., 2, 0] = -s[..., 1]
            matrix[..., 2, 1] = c[..., 1] * s[..., 2]
            matrix[..., 2, 2] = c[..., 1] * c[..., 2]
        elif order == "ZXY":
            matrix[..., 0, 0] = c[..., 0] * c[..., 2] - s[..., 0] * s[..., 1] * s[..., 2]
            matrix[..., 0, 1] = -c[..., 1] * s[..., 0]
            matrix[..., 0, 2] = c[..., 0] * s[..., 2] + c[..., 2] * s[..., 0] * s[..., 1]
            matrix[..., 1, 0] = c[..., 2] * s[..., 0] + c[..., 0] * s[..., 1] * s[..., 2]
            matrix[..., 1, 1] = c[..., 0] * c[..., 1]
            matrix[..., 1, 2] = s[..., 0] * s[..., 2] - c[..., 0] * c[..., 2] * s[..., 1]
            matrix[..., 2, 0] = -c[..., 1] * s[..., 2]
            matrix[..., 2, 1] = s[..., 1]
            matrix[..., 2, 2] = c[..., 1] * c[..., 2]
        elif order == "XZX":
            matrix[..., 0, 0] = c[..., 1]
            matrix[..., 0, 1] = -s[..., 1] * c[..., 2]
            matrix[..., 0, 2] = s[..., 1] * s[..., 2]
            matrix[..., 1, 0] = c[..., 0] * s[..., 1]
            matrix[..., 1, 1] = -s[..., 0] * s[..., 2] + c[..., 0] * c[..., 1] * c[..., 2]
            matrix[..., 1, 2] = -s[..., 0] * c[..., 2] - c[..., 0] * c[..., 1] * s[..., 2]
            matrix[..., 2, 0] = s[..., 0] * s[..., 1]
            matrix[..., 2, 1] = c[..., 0] * s[..., 2] + s[..., 0] * c[..., 1] * c[..., 2]
            matrix[..., 2, 2] = c[..., 0] * c[..., 2] - s[..., 0] * c[..., 1] * s[..., 2]
        elif order == "ZXZ":
            matrix[..., 0, 0] = c[..., 0] * c[..., 2] - s[..., 0] * c[..., 1] * s[..., 2]
            matrix[..., 0, 1] = -c[..., 0] * s[..., 2] - s[..., 0] * c[..., 1] * c[..., 2]
            matrix[..., 0, 2] = s[..., 0] * s[..., 1]
            matrix[..., 1, 0] = s[..., 0] * c[..., 2] + c[..., 0] * c[..., 1] * s[..., 2]
            matrix[..., 1, 1] = -s[..., 0] * s[..., 2] + c[..., 0] * c[..., 1] * c[..., 2]
            matrix[..., 1, 2] = -c[..., 0] * s[..., 1]
            matrix[..., 2, 0] = s[..., 1] * s[..., 2]
            matrix[..., 2, 1] = s[..., 1] * c[..., 2]
            matrix[..., 2, 2] = c[..., 1]
        else:
            matrix[..., 0, 0] = 1
            matrix[..., 1, 1] = 1
            matrix[..., 2, 2] = 1
            for i, char in enumerate(order):
                rot = matrix.new_empty(matrix.shape)
                if char == "X":
                    rot[..., 0, 0] = 1
                    rot[..., 0, 1] = 0
                    rot[..., 0, 2] = 0
                    rot[..., 1, 0] = 0
                    rot[..., 1, 1] = c[..., i]
                    rot[..., 1, 2] = -s[..., i]
                    rot[..., 2, 0] = 0
                    rot[..., 2, 1] = s[..., i]
                    rot[..., 2, 2] = c[..., i]
                elif char == "Y":
                    rot[..., 0, 0] = c[..., i]
                    rot[..., 0, 1] = 0
                    rot[..., 0, 2] = s[..., i]
                    rot[..., 1, 0] = 0
                    rot[..., 1, 1] = 1
                    rot[..., 1, 2] = 0
                    rot[..., 2, 0] = -s[..., i]
                    rot[..., 2, 1] = 0
                    rot[..., 2, 2] = c[..., i]
                elif char == "Z":
                    rot[..., 0, 0] = c[..., i]
                    rot[..., 0, 1] = -s[..., i]
                    rot[..., 0, 2] = 0
                    rot[..., 1, 0] = s[..., i]
                    rot[..., 1, 1] = c[..., i]
                    rot[..., 1, 2] = 0
                    rot[..., 2, 0] = 0
                    rot[..., 2, 1] = 0
                    rot[..., 2, 2] = 1
                matrix = rot if i == 0 else torch.bmm(matrix, rot)
    else:
        raise ValueError(
            f"Expected 'angles' to be scalar or tensor with last dimension size 3, got {N}"
        )
    return matrix


def euler_rotation_angles(matrix: Tensor, order: Optional[str] = None) -> Tensor:
    r"""Compute Euler angles from rotation matrix.

    TODO: Write test for this function and check for wich quadrant a rotation is in.
          See also https://github.com/BioMedIA/MIRTK/blob/77d3f391b49b0cee9e80da774fb074995fdf415f/Modules/Numerics/src/Matrix3x3.cc#L1217.

    Args:
        order: Order in which elemental rotations are composed. For example in 3D, "zxz" means
            that the first rotation occurs about z, the second about x, and the third rotation
            about z again. In 2D, this argument is ignored. Alternatively, strings of the form
            "Rz o Rx o Rz" can be given.

    Returns:
        Rotation angles in radians, where the first angle corresponds to the
            leftmost elemental rotation which is applied last.

    """
    if matrix.ndim < 2:
        raise ValueError("euler_rotation_angles() 'matrix' must be at least 2-dimensional")
    D = matrix.shape[-2]
    if D not in (2, 3) or matrix.shape[-1] not in (D, D + 1):
        raise ValueError(
            "euler_rotation_angles() 'matrix' must have shape (N, D, D) or (N, D, D + 1) where D=2 or D=3"
        )
    order = euler_rotation_order(order, ndim=D)
    det = matrix[..., :D, :D].detach().reshape(-1, D, D).det()
    if not det.abs().allclose(torch.tensor(1.0).to(det)):
        raise ValueError(
            "euler_rotation_angles() 'matrix' must be rotation matrix, i.e., matrix.det().abs() = 1"
        )
    if D == 2:
        angles = torch.acos(matrix[..., 0, 0])
    else:
        # https://en.wikipedia.org/wiki/Euler_angles#Rotation_matrix
        angles = matrix.new_empty(matrix.shape[:-2] + (D,))
        if order == "XZX":
            angles[..., 0] = torch.atan2(matrix[..., 0, 2], -matrix[..., 0, 1])
            angles[..., 1] = torch.acos(matrix[..., 0, 0])
            angles[..., 2] = torch.atan2(matrix[..., 2, 0], matrix[..., 1, 0])
        elif order == "ZXZ":
            angles[..., 0] = torch.atan2(matrix[..., 2, 0], matrix[..., 2, 1])
            angles[..., 1] = torch.acos(matrix[..., 2, 2])
            angles[..., 2] = torch.atan2(matrix[..., 0, 2], -matrix[..., 1, 2])
        else:
            raise NotImplementedError(f"euler_rotation_angles() order={order!r}")
    return angles


def euler_rotation_order(arg: Optional[str] = None, ndim: int = 3) -> str:
    r"""Standardize rotation order argument."""
    if arg is not None and not isinstance(arg, str):
        raise TypeError("euler_rotation_order() 'arg' must be str or None")
    if not isinstance(ndim, int):
        raise TypeError("euler_rotation_order() 'ndim' must be int")
    if ndim == 2:
        return "Z"
    if ndim != 3:
        raise NotImplementedError(f"euler_rotation_order() ndim={ndim}")
    order = "ZXZ" if arg is None else arg
    if re.match(r"^(R[xyz]|[XYZ])( o (R[xyz]|[XYZ]))*$", order):
        order = re.subn(r"R([xyz])", "\\1", order).replace(" o ", "")
    order = order.upper()
    if not re.match("^[XYZ][XYZ][XYZ]$", order):
        raise ValueError(f"euler_rotation_order() invalid argument '{arg}'")
    return order


def scaling_transform(
    scales: Union[Scalar, Array],
    homogeneous: bool = False,
    dtype: Optional[DType] = None,
    device: Optional[Device] = None,
) -> Tensor:
    r"""Scaling matrices.

    Args:
        scales: Tensor of anisotropic scaling factors of shape ``(..., D)``.
        homogeneous: Whether to return homogeneous transformation matrices.
        dtype: Data type of output matrix. If ``None``, use ``scales.dtype`` if it is
            a floating point type, and ``torch.float`` otherwise.
        device: Device on which to create matrix. If ``None``, use ``scales.device``.

    Returns:
        Tensor of square scaling matrices of shape ``(..., D, D)`` if ``homogeneous=False``, or tensor of
        homogeneous coordinate transformation matrices of shape ``(..., D, D + 1)`` otherwise.

    """
    scales_ = atleast_1d(scales, dtype=dtype, device=device)
    scales_ = as_float_tensor(scales_)
    D = scales_.shape[-1]
    J = tuple(range(D))
    matrix = scales_.new_zeros(scales_.shape[:-1] + (D, D + 1 if homogeneous else D))
    matrix[..., J, J] = scales_
    return matrix


def shear_matrix(
    angles: Union[Scalar, Array],
    homogeneous: bool = False,
    dtype: Optional[DType] = None,
    device: Optional[Device] = None,
) -> Tensor:
    r"""Shear matrices.

    Args:
        angles: Tensor of anisotropic shear angles in radians of shape ``(..., N)``,
            where ``N=(D * (D - 1)) / 2`` with ``D`` denoting the number of spatial dimensions.
        homogeneous: Whether to return homogeneous transformation matrices.
        dtype: Data type of output matrix. If ``None``, use ``angles.dtype`` if it is
            a floating point type, and ``torch.float`` otherwise.
        device: Device on which to create matrix. If ``None``, use ``angles.device``.

    Returns:
        Tensor of square scaling matrices of shape ``(..., D, D)`` if ``homogeneous=False``, or tensor of
        homogeneous coordinate transformation matrices of shape ``(..., D, D + 1)`` otherwise.

    """
    angles_ = atleast_1d(angles, dtype=dtype, device=device)
    angles_ = as_float_tensor(angles_)
    N = angles_.shape[-1]
    if N == 1:
        D = 2
    elif N == 3:
        D = 3
    elif N == 6:
        D = 4
    else:
        raise ValueError(
            "shear_matrix() 'angles' must have last dimension size 1 (2D), 3 (3D), or 6 (4D)"
        )
    J = tuple(range(D))
    K = torch.triu_indices(D, D, offset=1)
    matrix = angles_.new_zeros(angles_.shape[:-1] + (D, D + 1 if homogeneous else D))
    matrix[..., J, J] = 1
    matrix[..., K[0], K[1]] = torch.tan(angles_)
    return matrix


def translation(
    offset: Union[Scalar, Array],
    homogeneous: bool = False,
    dtype: Optional[DType] = None,
    device: Optional[Device] = None,
) -> Tensor:
    r"""Translation offsets / matrices.

    Args:
        offset: Translation vectors of shape ``(..., D)`` or ``(..., D, 1)``.
        homogeneous: Whether to return homogeneous transformation matrices.
        dtype: Data type of output matrix. If ``None``, use ``offset.dtype`` if it is
            a floating point type, and ``torch.float`` otherwise.
        device: Device on which to create matrix. If ``None``, use ``offset.device``.

    Returns:
        Homogeneous coordinate transformation matrices of shape ``(..., D, 1)`` if ``homogeneous=False``,
        or shape ``(..., D, D + 1)`` otherwise.

    """
    offset_ = atleast_1d(offset, dtype=dtype, device=device)
    offset_ = as_float_tensor(offset_)
    if homogeneous:
        if offset_.ndim > 1:
            offset_ = offset_.squeeze(-1)
        D = offset_.shape[-1]
        J = tuple(range(D))
        matrix = offset_.new_zeros(offset_.shape[:-1] + (D, D + 1))
        matrix[..., J, J] = 1
        matrix[..., D] = offset_
    elif offset_.ndim < 2 or offset_.shape[-1] != 1:
        matrix = offset_.unsqueeze(-1)
    else:
        matrix = offset_
    return matrix


def transform_points(transforms: Tensor, points: Tensor) -> Tensor:
    r"""Transform points by given homogeneous transformation."""
    return apply_transform(transforms, points, vectors=False)


def transform_vectors(transforms: Tensor, vectors: Tensor) -> Tensor:
    r"""Transform vectors by given homogeneous transformation."""
    return apply_transform(transforms, vectors, vectors=True)



r"""Image-to-Image transformer network (ITN)."""

import sys
from typing import Union

from deepali.networks.unet import UNet, UNetConfig as ImageTransformerConfig


class ImageTransformerNetwork(UNet):
    r"""Image transformer network."""

    def __init__(
        self,
        spatial_dims: int,
        in_channels: int = 1,
        config: Union[str, ImageTransformerConfig] = "default",
    ) -> None:
        if isinstance(config, str):
            config = itn_config(config)
        elif not isinstance(config, ImageTransformerConfig):
            raise TypeError(
                "ImageTransformerNetwork() 'config' must be str or ImageTransformerConfig"
            )
        if spatial_dims < 2 or spatial_dims > 3:
            raise ValueError("ImageTransformerNetwork() 'spatial_dims' must be 2 or 3")
        if in_channels < 1:
            raise ValueError("ImageTransformerNetwork() 'in_channels' must be positive")
        super().__init__(
            spatial_dims=spatial_dims,
            in_channels=in_channels,
            out_channels=in_channels,
            config=config,
        )


def itn_config(name: str = "default") -> ImageTransformerConfig:
    if name == "default":
        name = "miccai2019"
    config_factory = getattr(sys.modules[__name__], f"itn_config_{name.lower()}", None)
    if config_factory is None:
        raise ValueError(f"create_itn_model() has no 'config' named '{name}'")
    config = config_factory()
    assert isinstance(config, ImageTransformerConfig)
    return config


def itn_config_miccai2019() -> ImageTransformerConfig:
    config = ImageTransformerConfig()

    kernel_size = 3
    scale_factor = 2
    acti = "relu"
    bias = True
    norm = None

    config.encoder.num_channels = ((2, 4), (8, 8), (16, 16))
    config.encoder.conv_layer.kernel_size = kernel_size
    config.encoder.conv_layer.padding = kernel_size // 2
    config.encoder.conv_layer.bias = bias
    config.encoder.conv_layer.acti = acti
    config.encoder.conv_layer.norm = norm
    config.encoder.downsample.mode = "conv"
    config.encoder.downsample.factor = scale_factor
    config.encoder.downsample.kernel_size = scale_factor
    config.encoder.downsample.padding = 0

    config.decoder.num_channels = (16, (8, 8), (4, 2, 2))
    config.decoder.conv_layer.kernel_size = kernel_size
    config.encoder.conv_layer.padding = kernel_size // 2
    config.decoder.conv_layer.bias = bias
    config.decoder.conv_layer.acti = acti
    config.decoder.conv_layer.norm = norm
    config.decoder.upsample.mode = "deconv"
    config.decoder.upsample.factor = scale_factor
    config.decoder.upsample.kernel_size = scale_factor
    config.decoder.upsample.padding = 0
    config.decoder.join_mode = "add"

    config.output.kernel_size = kernel_size
    config.output.bias = bias
    config.output.acti = acti
    config.output.norm = norm

    return config
